"""Prepare to aspirate command request, result, and implementation models."""

from __future__ import annotations
from pydantic import BaseModel
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .pipetting_common import (
    PipetteIdMixin,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
)

if TYPE_CHECKING:
    from ..execution.pipetting import PipettingHandler

PrepareToAspirateCommandType = Literal["prepareToAspirate"]


class PrepareToAspirateParams(PipetteIdMixin):
    """Parameters required to prepare a specific pipette for aspiration."""

    pass


class PrepareToAspirateResult(BaseModel):
    """Result data from execution of an PrepareToAspirate command."""

    pass


class PrepareToAspirateImplementation(
    AbstractCommandImpl[
        PrepareToAspirateParams,
        PrepareToAspirateResult,
    ]
):
    """Prepare for aspirate command implementation."""

    def __init__(self, pipetting: PipettingHandler, **kwargs: object) -> None:
        self._pipetting_handler = pipetting

    async def execute(self, params: PrepareToAspirateParams) -> PrepareToAspirateResult:
        """Prepare the pipette to aspirate."""
        await self._pipetting_handler.prepare_for_aspirate(
            pipette_id=params.pipetteId,
        )

        return PrepareToAspirateResult()


class PrepareToAspirate(BaseCommand[PrepareToAspirateParams, PrepareToAspirateResult]):
    """Prepare for aspirate command model."""

    commandType: PrepareToAspirateCommandType = "prepareToAspirate"
    params: PrepareToAspirateParams
    result: Optional[PrepareToAspirateResult]

    _ImplementationCls: Type[
        PrepareToAspirateImplementation
    ] = PrepareToAspirateImplementation


class PrepareToAspirateCreate(BaseCommandCreate[PrepareToAspirateParams]):
    """Prepare for aspirate command creation request model."""

    commandType: PrepareToAspirateCommandType = "prepareToAspirate"
    params: PrepareToAspirateParams

    _CommandCls: Type[PrepareToAspirate] = PrepareToAspirate



"""Dispense command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from pydantic import Field

from ..types import DeckPoint
from .pipetting_common import (
    PipetteIdMixin,
    DispenseVolumeMixin,
    FlowRateMixin,
    WellLocationMixin,
    BaseLiquidHandlingResult,
    DestinationPositionResult,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler


DispenseCommandType = Literal["dispense"]


class DispenseParams(
    PipetteIdMixin, DispenseVolumeMixin, FlowRateMixin, WellLocationMixin
):
    """Payload required to dispense to a specific well."""

    pushOut: Optional[float] = Field(
        None,
        description="push the plunger a small amount farther than necessary for accurate low-volume dispensing",
    )


class DispenseResult(BaseLiquidHandlingResult, DestinationPositionResult):
    """Result data from the execution of a Dispense command."""

    pass


class DispenseImplementation(AbstractCommandImpl[DispenseParams, DispenseResult]):
    """Dispense command implementation."""

    def __init__(
        self, movement: MovementHandler, pipetting: PipettingHandler, **kwargs: object
    ) -> None:
        self._movement = movement
        self._pipetting = pipetting

    async def execute(self, params: DispenseParams) -> DispenseResult:
        """Move to and dispense to the requested well."""
        position = await self._movement.move_to_well(
            pipette_id=params.pipetteId,
            labware_id=params.labwareId,
            well_name=params.wellName,
            well_location=params.wellLocation,
        )
        volume = await self._pipetting.dispense_in_place(
            pipette_id=params.pipetteId,
            volume=params.volume,
            flow_rate=params.flowRate,
            push_out=params.pushOut,
        )

        return DispenseResult(
            volume=volume,
            position=DeckPoint(x=position.x, y=position.y, z=position.z),
        )


class Dispense(BaseCommand[DispenseParams, DispenseResult]):
    """Dispense command model."""

    commandType: DispenseCommandType = "dispense"
    params: DispenseParams
    result: Optional[DispenseResult]

    _ImplementationCls: Type[DispenseImplementation] = DispenseImplementation


class DispenseCreate(BaseCommandCreate[DispenseParams]):
    """Create dispense command request model."""

    commandType: DispenseCommandType = "dispense"
    params: DispenseParams

    _CommandCls: Type[Dispense] = Dispense



"""Validation file for addressable area reference checking functions."""

from opentrons.types import DeckSlotName


def is_waste_chute(addressable_area_name: str) -> bool:
    """Check if an addressable area is a Waste Chute."""
    return addressable_area_name in {
        "1ChannelWasteChute",
        "8ChannelWasteChute",
        "96ChannelWasteChute",
        "gripperWasteChute",
    }


def is_gripper_waste_chute(addressable_area_name: str) -> bool:
    """Check if an addressable area is a gripper-movement-compatible Waste Chute."""
    return addressable_area_name == "gripperWasteChute"


def is_drop_tip_waste_chute(addressable_area_name: str) -> bool:
    """Check if an addressable area is a Waste Chute compatible for dropping tips."""
    return addressable_area_name in {
        "1ChannelWasteChute",
        "8ChannelWasteChute",
        "96ChannelWasteChute",
    }


def is_trash(addressable_area_name: str) -> bool:
    """Check if an addressable area is a trash bin."""
    return addressable_area_name in {"movableTrash", "fixedTrash", "shortFixedTrash"}


def is_staging_slot(addressable_area_name: str) -> bool:
    """Check if an addressable area is a staging area slot."""
    return addressable_area_name in {"A4", "B4", "C4", "D4"}


def is_deck_slot(addressable_area_name: str) -> bool:
    """Check if an addressable area is a deck slot (including staging area slots)."""
    if is_staging_slot(addressable_area_name):
        return True
    try:
        DeckSlotName.from_primitive(addressable_area_name)
    except ValueError:
        return False
    return True



"""Command models to wait for target temperature of a Temperature Module."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from opentrons.protocol_engine.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

WaitForTemperatureCommandType = Literal["temperatureModule/waitForTemperature"]


class WaitForTemperatureParams(BaseModel):
    """Input parameters to wait for a Temperature Module's target temperature."""

    moduleId: str = Field(..., description="Unique ID of the Temperature Module.")
    celsius: Optional[float] = Field(
        None,
        description="Target temperature in °C. If not specified, will "
        "default to the module's target temperature. "
        "Specifying a celsius parameter other than the target temperature "
        "could lead to unpredictable behavior and hence is not "
        "recommended for use. This parameter can be removed in a "
        "future version without prior notice.",
    )


class WaitForTemperatureResult(BaseModel):
    """Result data from waiting for a Temperature Module's target temperature."""


class WaitForTemperatureImpl(
    AbstractCommandImpl[WaitForTemperatureParams, WaitForTemperatureResult]
):
    """Execution implementation of Temperature Module's wait for temperature command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: WaitForTemperatureParams
    ) -> WaitForTemperatureResult:
        """Wait for a Temperature Module's target temperature."""
        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        module_substate = self._state_view.modules.get_temperature_module_substate(
            module_id=params.moduleId
        )

        if params.celsius is None:
            # Raises error if no target temperature
            target_temp = module_substate.get_plate_target_temperature()
        else:
            target_temp = module_substate.validate_target_temperature(params.celsius)

        # Allow propagation of ModuleNotAttachedError.
        temp_hardware_module = self._equipment.get_module_hardware_api(
            module_substate.module_id
        )

        if temp_hardware_module is not None:
            await temp_hardware_module.await_temperature(
                awaiting_temperature=target_temp
            )
        return WaitForTemperatureResult()


class WaitForTemperature(
    BaseCommand[WaitForTemperatureParams, WaitForTemperatureResult]
):
    """A command to wait for a Temperature Module's target temperature."""

    commandType: WaitForTemperatureCommandType = "temperatureModule/waitForTemperature"
    params: WaitForTemperatureParams
    result: Optional[WaitForTemperatureResult]

    _ImplementationCls: Type[WaitForTemperatureImpl] = WaitForTemperatureImpl


class WaitForTemperatureCreate(BaseCommandCreate[WaitForTemperatureParams]):
    """A request to create a Temperature Module's wait for temperature command."""

    commandType: WaitForTemperatureCommandType = "temperatureModule/waitForTemperature"
    params: WaitForTemperatureParams

    _CommandCls: Type[WaitForTemperature] = WaitForTemperature



"""Union types of concrete command definitions."""

from typing import Union, TypeVar
from typing_extensions import Annotated

from pydantic import Field

from . import heater_shaker
from . import magnetic_module
from . import temperature_module
from . import thermocycler

from . import calibration

from .set_rail_lights import (
    SetRailLights,
    SetRailLightsCommandType,
    SetRailLightsCreate,
    SetRailLightsParams,
    SetRailLightsResult,
)

from .aspirate import (
    Aspirate,
    AspirateParams,
    AspirateCreate,
    AspirateResult,
    AspirateCommandType,
)

from .aspirate_in_place import (
    AspirateInPlace,
    AspirateInPlaceParams,
    AspirateInPlaceCreate,
    AspirateInPlaceResult,
    AspirateInPlaceCommandType,
)

from .comment import (
    Comment,
    CommentParams,
    CommentCreate,
    CommentResult,
    CommentCommandType,
)

from .custom import (
    Custom,
    CustomParams,
    CustomCreate,
    CustomResult,
    CustomCommandType,
)

from .dispense import (
    Dispense,
    DispenseParams,
    DispenseCreate,
    DispenseResult,
    DispenseCommandType,
)

from .dispense_in_place import (
    DispenseInPlace,
    DispenseInPlaceParams,
    DispenseInPlaceCreate,
    DispenseInPlaceResult,
    DispenseInPlaceCommandType,
)

from .drop_tip import (
    DropTip,
    DropTipParams,
    DropTipCreate,
    DropTipResult,
    DropTipCommandType,
)

from .drop_tip_in_place import (
    DropTipInPlace,
    DropTipInPlaceParams,
    DropTipInPlaceCreate,
    DropTipInPlaceResult,
    DropTipInPlaceCommandType,
)

from .home import (
    Home,
    HomeParams,
    HomeCreate,
    HomeResult,
    HomeCommandType,
)

from .load_labware import (
    LoadLabware,
    LoadLabwareParams,
    LoadLabwareCreate,
    LoadLabwareResult,
    LoadLabwareCommandType,
)

from .load_liquid import (
    LoadLiquid,
    LoadLiquidParams,
    LoadLiquidCreate,
    LoadLiquidResult,
    LoadLiquidCommandType,
)

from .load_module import (
    LoadModule,
    LoadModuleParams,
    LoadModuleCreate,
    LoadModuleResult,
    LoadModuleCommandType,
)

from .load_pipette import (
    LoadPipette,
    LoadPipetteParams,
    LoadPipetteCreate,
    LoadPipetteResult,
    LoadPipetteCommandType,
    LoadPipettePrivateResult,
)

from .move_labware import (
    MoveLabware,
    MoveLabwareParams,
    MoveLabwareCreate,
    MoveLabwareResult,
    MoveLabwareCommandType,
)

from .move_relative import (
    MoveRelative,
    MoveRelativeParams,
    MoveRelativeCreate,
    MoveRelativeResult,
    MoveRelativeCommandType,
)

from .move_to_coordinates import (
    MoveToCoordinates,
    MoveToCoordinatesParams,
    MoveToCoordinatesCreate,
    MoveToCoordinatesResult,
    MoveToCoordinatesCommandType,
)

from .move_to_well import (
    MoveToWell,
    MoveToWellParams,
    MoveToWellCreate,
    MoveToWellResult,
    MoveToWellCommandType,
)

from .move_to_addressable_area import (
    MoveToAddressableArea,
    MoveToAddressableAreaParams,
    MoveToAddressableAreaCreate,
    MoveToAddressableAreaResult,
    MoveToAddressableAreaCommandType,
)

from .move_to_addressable_area_for_drop_tip import (
    MoveToAddressableAreaForDropTip,
    MoveToAddressableAreaForDropTipParams,
    MoveToAddressableAreaForDropTipCreate,
    MoveToAddressableAreaForDropTipResult,
    MoveToAddressableAreaForDropTipCommandType,
)

from .wait_for_resume import (
    WaitForResume,
    WaitForResumeParams,
    WaitForResumeCreate,
    WaitForResumeResult,
    WaitForResumeCommandType,
)

from .wait_for_duration import (
    WaitForDuration,
    WaitForDurationParams,
    WaitForDurationCreate,
    WaitForDurationResult,
    WaitForDurationCommandType,
)

from .pick_up_tip import (
    PickUpTip,
    PickUpTipParams,
    PickUpTipCreate,
    PickUpTipResult,
    PickUpTipCommandType,
)

from .touch_tip import (
    TouchTip,
    TouchTipParams,
    TouchTipCreate,
    TouchTipResult,
    TouchTipCommandType,
)

from .save_position import (
    SavePosition,
    SavePositionParams,
    SavePositionCreate,
    SavePositionResult,
    SavePositionCommandType,
)

from .blow_out import (
    BlowOutParams,
    BlowOut,
    BlowOutCreate,
    BlowOutCommandType,
    BlowOutResult,
)

from .blow_out_in_place import (
    BlowOutInPlaceParams,
    BlowOutInPlace,
    BlowOutInPlaceCreate,
    BlowOutInPlaceCommandType,
    BlowOutInPlaceResult,
)

from .set_status_bar import (
    SetStatusBar,
    SetStatusBarParams,
    SetStatusBarCreate,
    SetStatusBarResult,
    SetStatusBarCommandType,
)

from .retract_axis import (
    RetractAxis,
    RetractAxisParams,
    RetractAxisCreate,
    RetractAxisResult,
    RetractAxisCommandType,
)

from .configure_for_volume import (
    ConfigureForVolume,
    ConfigureForVolumeParams,
    ConfigureForVolumeCreate,
    ConfigureForVolumeResult,
    ConfigureForVolumeCommandType,
    ConfigureForVolumePrivateResult,
)

from .prepare_to_aspirate import (
    PrepareToAspirate,
    PrepareToAspirateParams,
    PrepareToAspirateCreate,
    PrepareToAspirateResult,
    PrepareToAspirateCommandType,
)

from .configure_nozzle_layout import (
    ConfigureNozzleLayout,
    ConfigureNozzleLayoutCreate,
    ConfigureNozzleLayoutParams,
    ConfigureNozzleLayoutResult,
    ConfigureNozzleLayoutCommandType,
    ConfigureNozzleLayoutPrivateResult,
)

from .verify_tip_presence import (
    VerifyTipPresence,
    VerifyTipPresenceCreate,
    VerifyTipPresenceParams,
    VerifyTipPresenceResult,
    VerifyTipPresenceCommandType,
)

from .get_tip_presence import (
    GetTipPresence,
    GetTipPresenceCreate,
    GetTipPresenceParams,
    GetTipPresenceResult,
    GetTipPresenceCommandType,
)

Command = Annotated[
    Union[
        Aspirate,
        AspirateInPlace,
        Comment,
        Custom,
        Dispense,
        DispenseInPlace,
        BlowOut,
        BlowOutInPlace,
        ConfigureForVolume,
        ConfigureNozzleLayout,
        DropTip,
        DropTipInPlace,
        Home,
        RetractAxis,
        LoadLabware,
        LoadLiquid,
        LoadModule,
        LoadPipette,
        MoveLabware,
        MoveRelative,
        MoveToCoordinates,
        MoveToWell,
        MoveToAddressableArea,
        MoveToAddressableAreaForDropTip,
        PrepareToAspirate,
        WaitForResume,
        WaitForDuration,
        PickUpTip,
        SavePosition,
        SetRailLights,
        TouchTip,
        SetStatusBar,
        VerifyTipPresence,
        GetTipPresence,
        heater_shaker.WaitForTemperature,
        heater_shaker.SetTargetTemperature,
        heater_shaker.DeactivateHeater,
        heater_shaker.SetAndWaitForShakeSpeed,
        heater_shaker.DeactivateShaker,
        heater_shaker.OpenLabwareLatch,
        heater_shaker.CloseLabwareLatch,
        magnetic_module.Disengage,
        magnetic_module.Engage,
        temperature_module.SetTargetTemperature,
        temperature_module.WaitForTemperature,
        temperature_module.DeactivateTemperature,
        thermocycler.SetTargetBlockTemperature,
        thermocycler.WaitForBlockTemperature,
        thermocycler.SetTargetLidTemperature,
        thermocycler.WaitForLidTemperature,
        thermocycler.DeactivateBlock,
        thermocycler.DeactivateLid,
        thermocycler.OpenLid,
        thermocycler.CloseLid,
        thermocycler.RunProfile,
        calibration.CalibrateGripper,
        calibration.CalibratePipette,
        calibration.CalibrateModule,
        calibration.MoveToMaintenancePosition,
    ],
    Field(discriminator="commandType"),
]

CommandParams = Union[
    AspirateParams,
    AspirateInPlaceParams,
    CommentParams,
    ConfigureForVolumeParams,
    ConfigureNozzleLayoutParams,
    CustomParams,
    DispenseParams,
    DispenseInPlaceParams,
    BlowOutParams,
    BlowOutInPlaceParams,
    DropTipParams,
    DropTipInPlaceParams,
    HomeParams,
    RetractAxisParams,
    LoadLabwareParams,
    LoadLiquidParams,
    LoadModuleParams,
    LoadPipetteParams,
    MoveLabwareParams,
    MoveRelativeParams,
    MoveToCoordinatesParams,
    MoveToWellParams,
    MoveToAddressableAreaParams,
    MoveToAddressableAreaForDropTipParams,
    PrepareToAspirateParams,
    WaitForResumeParams,
    WaitForDurationParams,
    PickUpTipParams,
    SavePositionParams,
    SetRailLightsParams,
    TouchTipParams,
    SetStatusBarParams,
    VerifyTipPresenceParams,
    GetTipPresenceParams,
    heater_shaker.WaitForTemperatureParams,
    heater_shaker.SetTargetTemperatureParams,
    heater_shaker.DeactivateHeaterParams,
    heater_shaker.SetAndWaitForShakeSpeedParams,
    heater_shaker.DeactivateShakerParams,
    heater_shaker.OpenLabwareLatchParams,
    heater_shaker.CloseLabwareLatchParams,
    magnetic_module.DisengageParams,
    magnetic_module.EngageParams,
    temperature_module.SetTargetTemperatureParams,
    temperature_module.WaitForTemperatureParams,
    temperature_module.DeactivateTemperatureParams,
    thermocycler.SetTargetBlockTemperatureParams,
    thermocycler.WaitForBlockTemperatureParams,
    thermocycler.SetTargetLidTemperatureParams,
    thermocycler.WaitForLidTemperatureParams,
    thermocycler.DeactivateBlockParams,
    thermocycler.DeactivateLidParams,
    thermocycler.OpenLidParams,
    thermocycler.CloseLidParams,
    thermocycler.RunProfileParams,
    thermocycler.RunProfileStepParams,
    calibration.CalibrateGripperParams,
    calibration.CalibratePipetteParams,
    calibration.CalibrateModuleParams,
    calibration.MoveToMaintenancePositionParams,
]

CommandType = Union[
    AspirateCommandType,
    AspirateInPlaceCommandType,
    CommentCommandType,
    ConfigureForVolumeCommandType,
    ConfigureNozzleLayoutCommandType,
    CustomCommandType,
    DispenseCommandType,
    DispenseInPlaceCommandType,
    BlowOutCommandType,
    BlowOutInPlaceCommandType,
    DropTipCommandType,
    DropTipInPlaceCommandType,
    HomeCommandType,
    RetractAxisCommandType,
    LoadLabwareCommandType,
    LoadLiquidCommandType,
    LoadModuleCommandType,
    LoadPipetteCommandType,
    MoveLabwareCommandType,
    MoveRelativeCommandType,
    MoveToCoordinatesCommandType,
    MoveToWellCommandType,
    MoveToAddressableAreaCommandType,
    MoveToAddressableAreaForDropTipCommandType,
    PrepareToAspirateCommandType,
    WaitForResumeCommandType,
    WaitForDurationCommandType,
    PickUpTipCommandType,
    SavePositionCommandType,
    SetRailLightsCommandType,
    TouchTipCommandType,
    SetStatusBarCommandType,
    VerifyTipPresenceCommandType,
    GetTipPresenceCommandType,
    heater_shaker.WaitForTemperatureCommandType,
    heater_shaker.SetTargetTemperatureCommandType,
    heater_shaker.DeactivateHeaterCommandType,
    heater_shaker.SetAndWaitForShakeSpeedCommandType,
    heater_shaker.DeactivateShakerCommandType,
    heater_shaker.OpenLabwareLatchCommandType,
    heater_shaker.CloseLabwareLatchCommandType,
    magnetic_module.DisengageCommandType,
    magnetic_module.EngageCommandType,
    temperature_module.SetTargetTemperatureCommandType,
    temperature_module.WaitForTemperatureCommandType,
    temperature_module.DeactivateTemperatureCommandType,
    thermocycler.SetTargetBlockTemperatureCommandType,
    thermocycler.WaitForBlockTemperatureCommandType,
    thermocycler.SetTargetLidTemperatureCommandType,
    thermocycler.WaitForLidTemperatureCommandType,
    thermocycler.DeactivateBlockCommandType,
    thermocycler.DeactivateLidCommandType,
    thermocycler.OpenLidCommandType,
    thermocycler.CloseLidCommandType,
    thermocycler.RunProfileCommandType,
    calibration.CalibrateGripperCommandType,
    calibration.CalibratePipetteCommandType,
    calibration.CalibrateModuleCommandType,
    calibration.MoveToMaintenancePositionCommandType,
]

CommandCreate = Annotated[
    Union[
        AspirateCreate,
        AspirateInPlaceCreate,
        CommentCreate,
        ConfigureForVolumeCreate,
        ConfigureNozzleLayoutCreate,
        CustomCreate,
        DispenseCreate,
        DispenseInPlaceCreate,
        BlowOutCreate,
        BlowOutInPlaceCreate,
        DropTipCreate,
        DropTipInPlaceCreate,
        HomeCreate,
        RetractAxisCreate,
        LoadLabwareCreate,
        LoadLiquidCreate,
        LoadModuleCreate,
        LoadPipetteCreate,
        MoveLabwareCreate,
        MoveRelativeCreate,
        MoveToCoordinatesCreate,
        MoveToWellCreate,
        MoveToAddressableAreaCreate,
        MoveToAddressableAreaForDropTipCreate,
        PrepareToAspirateCreate,
        WaitForResumeCreate,
        WaitForDurationCreate,
        PickUpTipCreate,
        SavePositionCreate,
        SetRailLightsCreate,
        TouchTipCreate,
        SetStatusBarCreate,
        VerifyTipPresenceCreate,
        GetTipPresenceCreate,
        heater_shaker.WaitForTemperatureCreate,
        heater_shaker.SetTargetTemperatureCreate,
        heater_shaker.DeactivateHeaterCreate,
        heater_shaker.SetAndWaitForShakeSpeedCreate,
        heater_shaker.DeactivateShakerCreate,
        heater_shaker.OpenLabwareLatchCreate,
        heater_shaker.CloseLabwareLatchCreate,
        magnetic_module.DisengageCreate,
        magnetic_module.EngageCreate,
        temperature_module.SetTargetTemperatureCreate,
        temperature_module.WaitForTemperatureCreate,
        temperature_module.DeactivateTemperatureCreate,
        thermocycler.SetTargetBlockTemperatureCreate,
        thermocycler.WaitForBlockTemperatureCreate,
        thermocycler.SetTargetLidTemperatureCreate,
        thermocycler.WaitForLidTemperatureCreate,
        thermocycler.DeactivateBlockCreate,
        thermocycler.DeactivateLidCreate,
        thermocycler.OpenLidCreate,
        thermocycler.CloseLidCreate,
        thermocycler.RunProfileCreate,
        calibration.CalibrateGripperCreate,
        calibration.CalibratePipetteCreate,
        calibration.CalibrateModuleCreate,
        calibration.MoveToMaintenancePositionCreate,
    ],
    Field(discriminator="commandType"),
]

CommandResult = Union[
    AspirateResult,
    AspirateInPlaceResult,
    CommentResult,
    ConfigureForVolumeResult,
    ConfigureNozzleLayoutResult,
    CustomResult,
    DispenseResult,
    DispenseInPlaceResult,
    BlowOutResult,
    BlowOutInPlaceResult,
    DropTipResult,
    DropTipInPlaceResult,
    HomeResult,
    RetractAxisResult,
    LoadLabwareResult,
    LoadLiquidResult,
    LoadModuleResult,
    LoadPipetteResult,
    MoveLabwareResult,
    MoveRelativeResult,
    MoveToCoordinatesResult,
    MoveToWellResult,
    MoveToAddressableAreaResult,
    MoveToAddressableAreaForDropTipResult,
    PrepareToAspirateResult,
    WaitForResumeResult,
    WaitForDurationResult,
    PickUpTipResult,
    SavePositionResult,
    SetRailLightsResult,
    TouchTipResult,
    SetStatusBarResult,
    VerifyTipPresenceResult,
    GetTipPresenceResult,
    heater_shaker.WaitForTemperatureResult,
    heater_shaker.SetTargetTemperatureResult,
    heater_shaker.DeactivateHeaterResult,
    heater_shaker.SetAndWaitForShakeSpeedResult,
    heater_shaker.DeactivateShakerResult,
    heater_shaker.OpenLabwareLatchResult,
    heater_shaker.CloseLabwareLatchResult,
    magnetic_module.DisengageResult,
    magnetic_module.EngageResult,
    temperature_module.SetTargetTemperatureResult,
    temperature_module.WaitForTemperatureResult,
    temperature_module.DeactivateTemperatureResult,
    thermocycler.SetTargetBlockTemperatureResult,
    thermocycler.WaitForBlockTemperatureResult,
    thermocycler.SetTargetLidTemperatureResult,
    thermocycler.WaitForLidTemperatureResult,
    thermocycler.DeactivateBlockResult,
    thermocycler.DeactivateLidResult,
    thermocycler.OpenLidResult,
    thermocycler.CloseLidResult,
    thermocycler.RunProfileResult,
    calibration.CalibrateGripperResult,
    calibration.CalibratePipetteResult,
    calibration.CalibrateModuleResult,
    calibration.MoveToMaintenancePositionResult,
]

CommandPrivateResult = Union[
    None,
    LoadPipettePrivateResult,
    ConfigureForVolumePrivateResult,
    ConfigureNozzleLayoutPrivateResult,
]

CommandT = TypeVar("CommandT", bound=Command)



from .controller import Controller
from .simulator import Simulator

# only expose the ot2 interfaces in __init__ so everything works if opentrons_hardware
# is not present

__all__ = ["Controller", "Simulator"]



""" opentrons.system: Functions and data for interacting with the OT2 system.

The functions in the modules contained in this package do things like bridge
Python calls with shell invocations of nmcli or other system utilities. They
are in general unsafe to call when not running on an Opentrons robot.
"""



"""Move to coordinates command request, result, and implementation models."""
from __future__ import annotations

from pydantic import Field
from typing import Optional, Type, TYPE_CHECKING
from typing_extensions import Literal

from ..types import DeckPoint
from .pipetting_common import PipetteIdMixin, MovementMixin, DestinationPositionResult
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..execution import MovementHandler


MoveToCoordinatesCommandType = Literal["moveToCoordinates"]


class MoveToCoordinatesParams(PipetteIdMixin, MovementMixin):
    """Payload required to move a pipette to coordinates."""

    coordinates: DeckPoint = Field(
        ...,
        description="X, Y and Z coordinates in mm from deck's origin location (left-front-bottom corner of work space)",
    )


class MoveToCoordinatesResult(DestinationPositionResult):
    """Result data from the execution of a MoveToCoordinates command."""

    pass


class MoveToCoordinatesImplementation(
    AbstractCommandImpl[MoveToCoordinatesParams, MoveToCoordinatesResult]
):
    """Move to coordinates command implementation."""

    def __init__(
        self,
        movement: MovementHandler,
        **kwargs: object,
    ) -> None:
        self._movement = movement

    async def execute(self, params: MoveToCoordinatesParams) -> MoveToCoordinatesResult:
        """Move the requested pipette to the requested coordinates."""
        x, y, z = await self._movement.move_to_coordinates(
            pipette_id=params.pipetteId,
            deck_coordinates=params.coordinates,
            direct=params.forceDirect,
            additional_min_travel_z=params.minimumZHeight,
            speed=params.speed,
        )

        return MoveToCoordinatesResult(position=DeckPoint(x=x, y=y, z=z))


class MoveToCoordinates(BaseCommand[MoveToCoordinatesParams, MoveToCoordinatesResult]):
    """Move to well command model."""

    commandType: MoveToCoordinatesCommandType = "moveToCoordinates"
    params: MoveToCoordinatesParams
    result: Optional[MoveToCoordinatesResult]

    _ImplementationCls: Type[
        MoveToCoordinatesImplementation
    ] = MoveToCoordinatesImplementation


class MoveToCoordinatesCreate(BaseCommandCreate[MoveToCoordinatesParams]):
    """Move to coordinates command creation request model."""

    commandType: MoveToCoordinatesCommandType = "moveToCoordinates"
    params: MoveToCoordinatesParams

    _CommandCls: Type[MoveToCoordinates] = MoveToCoordinates



from opentrons.types import DeckSlotName


class TrashBin:
    """Represents a Flex or OT-2 trash bin.

    See :py:meth:`.ProtocolContext.load_trash_bin`.
    """

    def __init__(self, location: DeckSlotName, addressable_area_name: str) -> None:
        self._location = location
        self._addressable_area_name = addressable_area_name

    @property
    def location(self) -> DeckSlotName:
        """Location of the trash bin.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return self._location

    @property
    def area_name(self) -> str:
        """Addressable area name of the trash bin.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return self._addressable_area_name



import json
import logging
from pydantic import ValidationError
from typing import Optional, Union, no_type_check
from dataclasses import asdict

from opentrons import config, types


from .. import file_operators as io, types as local_types
from opentrons.util.helpers import utc_now

from .models import v1

log = logging.getLogger(__name__)
# Delete Pipette Offset Calibrations


@no_type_check
def delete_pipette_offset_file(pipette: str, mount: types.Mount) -> None:
    """
    Delete pipette offset file based on mount and pipette serial number

    :param pipette: pipette serial number
    :param mount: pipette mount
    """
    offset_dir = config.get_opentrons_path("pipette_calibration_dir")
    offset_path = offset_dir / mount.name.lower() / f"{pipette}.json"
    io.delete_file(offset_path)


def clear_pipette_offset_calibrations() -> None:
    """
    Delete all pipette offset calibration files.
    """

    offset_dir = config.get_opentrons_path("pipette_calibration_dir")
    io._remove_json_files_in_directories(offset_dir)


# Save Pipette Offset Calibrations


@no_type_check
def save_pipette_calibration(
    offset: types.Point,
    pip_id: str,
    mount: types.Mount,
    cal_status: Optional[
        Union[local_types.CalibrationStatus, v1.CalibrationStatus]
    ] = None,
) -> None:
    pip_dir = config.get_opentrons_path("pipette_calibration_dir") / mount.name.lower()

    if isinstance(cal_status, local_types.CalibrationStatus):
        cal_status_model = v1.CalibrationStatus(**asdict(cal_status))
    elif isinstance(cal_status, v1.CalibrationStatus):
        cal_status_model = cal_status
    else:
        cal_status_model = v1.CalibrationStatus()

    pipette_calibration = v1.InstrumentOffsetModel(
        offset=offset,
        lastModified=utc_now(),
        source=local_types.SourceType.user,
        status=cal_status_model,
    )
    io.save_to_file(pip_dir, pip_id, pipette_calibration)


# Get Pipette Offset Calibrations


@no_type_check
def get_pipette_offset(
    pipette_id: str, mount: types.Mount
) -> Optional[v1.InstrumentOffsetModel]:
    try:
        pipette_calibration_filepath = (
            config.get_opentrons_path("pipette_calibration_dir")
            / mount.name.lower()
            / f"{pipette_id}.json"
        )
        return v1.InstrumentOffsetModel(
            **io.read_cal_file(pipette_calibration_filepath)
        )
    except FileNotFoundError:
        log.debug(f"Calibrations for {pipette_id} on {mount} does not exist.")
        return None
    except (json.JSONDecodeError, ValidationError):
        log.warning(
            f"Malformed calibrations for {pipette_id} on {mount}. Please factory reset your calibrations.",
            exc_info=True,
        )
        return None



# NOTE(mc, 2022-12-05): this module is deprecated;
# do not add to it unless necessary and do not use in new code
import functools
import logging
from collections import UserDict
from typing import Dict, Optional, List, Union, Mapping
from typing_extensions import Protocol, Final

from opentrons_shared_data.deck import load as load_deck

from opentrons_shared_data.deck.dev_types import SlotDefV3
from opentrons_shared_data.labware.dev_types import LabwareUri

from opentrons.hardware_control.modules.types import ModuleType
from opentrons.motion_planning import deck_conflict
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.types import (
    DeckLocation,
    Location,
    Mount,
    Point,
    DeckSlotName,
    StagingSlotName,
)

from opentrons.protocol_api.core.labware import AbstractLabware
from opentrons.protocol_api.deck import CalibrationPosition
from opentrons.protocol_api.labware import load as load_lw, Labware

from .legacy_labware_core import LegacyLabwareCore
from .module_geometry import ModuleGeometry, HeaterShakerGeometry, ThermocyclerGeometry


_log = logging.getLogger(__name__)


# Amount of slots in a single deck row
ROW_LENGTH = 3
FIXED_TRASH_ID = "fixedTrash"


DEFAULT_LEGACY_DECK_DEFINITION_VERSION: Final = 3


class DeckItem(Protocol):
    @property
    def highest_z(self) -> float:
        ...

    @property
    def load_name(self) -> str:
        ...


class Deck(UserDict):  # type: ignore[type-arg]
    data: Dict[int, Optional[DeckItem]]

    def __init__(self, deck_type: str) -> None:
        super().__init__()
        self._definition = load_deck(
            name=deck_type, version=DEFAULT_LEGACY_DECK_DEFINITION_VERSION
        )
        self._positions = {}
        for slot in self._definition["locations"]["orderedSlots"]:
            self.data[int(slot["id"])] = None
            self._positions[int(slot["id"])] = Point(*slot["position"])
        self._highest_z = 0.0
        self._load_fixtures()
        self._thermocycler_present = False

    def _load_fixtures(self) -> None:
        for f in self._definition["locations"]["fixtures"]:
            slot_name = self._check_name(f["slot"])  # type: ignore
            # TODO(mc, 2022-06-15): this loads the fixed trash as an instance of
            # `opentrons.protocol_api.labware.Labware`
            # However, all other labware will be added to the `Deck` as instances of
            # `opentrons.protocol_api.core.labware.AbstractLabware`
            # And modules will be added as instances of
            # `opentrons.protocols.geometry.module_geometry.ModuleGeometry`
            # This mix of public and private interfaces as members of a public
            # `Deck` interface is confusing and should be resolved.
            # If `Deck` is public, all items in a `Deck` should be
            # instances of other public APIs.
            loaded_f = load_lw(
                f["labware"], self.position_for(slot_name)  # type: ignore
            )
            self.__setitem__(slot_name, loaded_f)

    @staticmethod
    @functools.lru_cache(20)
    def _assure_int(key: object) -> int:
        if isinstance(key, str):
            return int(key)
        elif isinstance(key, int):
            return key
        else:
            raise TypeError(type(key))

    def _check_name(self, key: object) -> int:
        should_raise = False
        try:
            key_int = Deck._assure_int(key)
        except Exception:
            _log.exception("Bad slot name: {}".format(key))
            should_raise = True
        should_raise = should_raise or key_int not in self.data
        if should_raise:
            raise ValueError("Unknown slot: {}".format(key))
        else:
            return key_int

    @staticmethod
    def _map_to_conflict_checker_item(item: DeckItem) -> deck_conflict.DeckItem:
        name_for_errors = item.load_name

        # We have to account for both Labware and AbstractLabware because this class
        # inappropriately contains both internal and customer-facing types.
        # See todo comment in self._load_fixtures().
        if isinstance(item, Labware):
            is_fixed_trash = "fixedTrash" in item.quirks
            return deck_conflict.Labware(
                highest_z=item.highest_z,
                name_for_errors=name_for_errors,
                # TODO(mm, 2023-02-16): Refactor item.uri to return LabwareUri.
                uri=LabwareUri(item.uri),
                is_fixed_trash=is_fixed_trash,
            )

        elif isinstance(item, AbstractLabware):
            is_fixed_trash = "fixedTrash" in item.get_quirks()
            return deck_conflict.Labware(
                highest_z=item.highest_z,
                name_for_errors=name_for_errors,
                # TODO(mm, 2023-02-16): Refactor item.uri to return LabwareUri.
                uri=LabwareUri(item.get_uri()),
                is_fixed_trash=is_fixed_trash,
            )

        elif isinstance(item, HeaterShakerGeometry):
            return deck_conflict.HeaterShakerModule(
                highest_z_including_labware=item.highest_z,
                name_for_errors=name_for_errors,
            )

        elif isinstance(item, ThermocyclerGeometry):
            return deck_conflict.ThermocyclerModule(
                highest_z_including_labware=item.highest_z,
                name_for_errors=name_for_errors,
                is_semi_configuration=item.is_semi_configuration,
            )

        elif isinstance(item, ModuleGeometry):
            return deck_conflict.OtherModule(
                highest_z_including_labware=item.highest_z,
                name_for_errors=name_for_errors,
            )

        else:
            assert False, f"Deck item {item} has an unknown type."

    def __getitem__(self, key: DeckLocation) -> Optional[DeckItem]:
        return self.data[self._check_name(key)]

    def __delitem__(self, key: DeckLocation) -> None:
        checked_key = self._check_name(key)
        old = self.data[checked_key]
        self.data[checked_key] = None
        if old:
            self.recalculate_high_z()
            # Update the thermocycler present member
            self._thermocycler_present = any(
                isinstance(item, ThermocyclerGeometry) for item in self.data.values()
            )

    def __setitem__(self, key: DeckLocation, val: DeckItem) -> None:
        slot_key_int = self._check_name(key)
        existing_items: Mapping[
            Union[DeckSlotName, StagingSlotName], deck_conflict.DeckItem
        ] = {
            DeckSlotName.from_primitive(slot): self._map_to_conflict_checker_item(item)
            for slot, item in self.data.items()
            if item is not None
        }

        # will raise DeckConflictError if items conflict
        deck_conflict.check(
            existing_items=existing_items,
            new_location=DeckSlotName.from_primitive(slot_key_int),
            new_item=self._map_to_conflict_checker_item(val),
            robot_type=self._definition["robot"]["model"],
        )

        self.data[slot_key_int] = val
        self.recalculate_high_z()
        self._thermocycler_present = any(
            isinstance(item, ThermocyclerGeometry) for item in self.data.values()
        )

    def __contains__(self, key: object) -> bool:
        try:
            key_int = self._check_name(key)
        except ValueError:
            return False
        return key_int in self.data

    def is_edge_move_unsafe(self, mount: Mount, target: "Labware") -> bool:
        """
        Check if slot next to target labware contains a module. Only relevant
        depending on the mount you are using and the column you are moving
        to inside of the labware.
        """
        slot = LabwareLike(target).first_parent()
        if not slot:
            return False
        if mount is Mount.RIGHT:
            other_labware = self.left_of(slot)
        else:
            other_labware = self.right_of(slot)

        return isinstance(other_labware, ModuleGeometry)

    def right_of(self, slot: str) -> Optional[DeckItem]:
        if int(slot) % ROW_LENGTH == 0:
            # We know we're at the right-most edge
            # of the given row
            return None
        else:
            idx = int(slot) + 1
            return self[str(idx)]

    def left_of(self, slot: str) -> Optional[DeckItem]:
        if int(slot) - 1 % ROW_LENGTH == 0:
            # We know we're at the left-most edge
            # of the given row
            return None
        idx = int(slot) - 1
        if idx < 1:
            return None
        return self[str(idx)]

    def position_for(self, key: DeckLocation) -> Location:
        key_int = self._check_name(key)
        return Location(self._positions[key_int], str(key))

    def recalculate_high_z(self) -> None:
        self._highest_z = 0.0
        for item in [lw for lw in self.data.values() if lw]:
            self._highest_z = max(item.highest_z, self._highest_z)

    def get_slot_definition(self, slot_name: str) -> SlotDefV3:
        slots = self._definition["locations"]["orderedSlots"]
        slot_def = next((slot for slot in slots if slot["id"] == slot_name), None)
        if not slot_def:
            slot_ids = [slot["id"] for slot in slots]
            raise ValueError(
                f"slot {slot_name} could not be found,"
                f"valid deck slots are: {slot_ids}"
            )
        return slot_def

    def get_slot_center(self, slot_name: str) -> Point:
        defn = self.get_slot_definition(slot_name)
        return Point(
            defn["position"][0] + defn["boundingBox"]["xDimension"] / 2,
            defn["position"][1] + defn["boundingBox"]["yDimension"] / 2,
            defn["position"][2] + defn["boundingBox"]["zDimension"] / 2,
        )

    def resolve_module_location(
        self, module_type: ModuleType, location: Optional[DeckLocation]
    ) -> DeckLocation:
        dn_from_type = {
            ModuleType.MAGNETIC: "Magnetic Module",
            ModuleType.THERMOCYCLER: "Thermocycler",
            ModuleType.TEMPERATURE: "Temperature Module",
            ModuleType.HEATER_SHAKER: "Heater-Shaker",
        }
        if isinstance(location, str) or isinstance(location, int):
            slot_def = self.get_slot_definition(str(location))
            compatible_modules = slot_def["compatibleModuleTypes"]
            if module_type.value in compatible_modules:
                return location
            else:
                raise ValueError(
                    f"A {dn_from_type[module_type]} cannot be loaded"
                    f" into slot {location}"
                )
        else:
            valid_slots = [
                slot["id"]
                for slot in self.slots
                if module_type.value in slot["compatibleModuleTypes"]
            ]
            if len(valid_slots) == 1:
                return valid_slots[0]
            elif not valid_slots:
                raise ValueError(
                    "A {dn_from_type[module_type]} cannot be used with this deck"
                )
            else:
                raise ValueError(
                    f"{dn_from_type[module_type]}s do not have default"
                    " location, you must specify a slot"
                )

    @property
    def highest_z(self) -> float:
        """Return the tallest known point on the deck."""
        return self._highest_z

    @property
    def slots(self) -> List[SlotDefV3]:
        """Return the definition of the loaded robot deck."""
        return self._definition["locations"]["orderedSlots"]

    @property
    def calibration_positions(self) -> List[CalibrationPosition]:
        raw_positions = self._definition["locations"]["calibrationPoints"]
        return [
            CalibrationPosition(
                id=raw_position["id"],
                displayName=raw_position["displayName"],
                position=(
                    raw_position["position"][0],
                    raw_position["position"][1],
                    raw_position["position"][2],
                ),
            )
            for raw_position in raw_positions
        ]

    def get_calibration_position(self, id: str) -> CalibrationPosition:
        calibration_position = next(
            (pos for pos in self.calibration_positions if pos.id == id), None
        )
        if not calibration_position:
            pos_ids = [pos.id for pos in self.calibration_positions]
            raise ValueError(
                f"calibration position {id} "
                "could not be found, "
                f"valid calibration position ids are: {pos_ids}"
            )
        return calibration_position

    def get_fixed_trash(self) -> Optional[Union[Labware, LegacyLabwareCore]]:
        fixtures = self._definition["locations"]["fixtures"]
        ft = next((f for f in fixtures if f["id"] == FIXED_TRASH_ID), None)

        # NOTE(mc, 2022-12-06): type ignore below because this `Deck` is typed
        # as contiaining `DeckItem` values, but the contents of slot 12
        # will always be limited to either a `Labware` or `LegacyLabwareCore` object
        return self.data[self._check_name(ft.get("slot"))] if ft else None  # type: ignore[return-value]

    def get_non_fixture_slots(self) -> List[int]:
        fixtures = self._definition["locations"]["fixtures"]
        fixture_slots = {
            self._check_name(f.get("slot")) for f in fixtures if f.get("slot")
        }
        return [s for s in self.data.keys() if s not in fixture_slots]

    @property
    def thermocycler_present(self) -> bool:
        """Is a thermocycler present on the deck."""
        return self._thermocycler_present



import json
import logging
import os
from functools import lru_cache
from pathlib import Path
from opentrons.hardware_control.modules.types import ModuleType
from opentrons.hardware_control.types import OT3Mount
from pydantic import ValidationError
from typing import List, Optional, Union, no_type_check
from dataclasses import asdict

from opentrons import config, types


from .. import file_operators as io, types as local_types
from opentrons.util.helpers import utc_now

from .models import v1

log = logging.getLogger(__name__)
# Delete Module Offset Calibrations


@no_type_check
def delete_module_offset_file(module_id: str) -> None:
    """
    Delete module offset file for the given module id.

    :param module: module id serial number
    """
    offset_dir = config.get_opentrons_path("module_calibration_dir")
    offset_path = offset_dir / f"{module_id}.json"
    io.delete_file(offset_path)
    # something changed, clear lru cache
    get_module_offset.cache_clear()
    load_all_module_offsets.cache_clear()


def clear_module_offset_calibrations() -> None:
    """
    Delete all module offset calibration files.
    """

    offset_dir = config.get_opentrons_path("module_calibration_dir")
    io._remove_json_files_in_directories(offset_dir)
    # something changed, clear lru cache
    get_module_offset.cache_clear()
    load_all_module_offsets.cache_clear()


# Save Module Offset Calibrations


@no_type_check
def save_module_calibration(
    offset: types.Point,
    mount: OT3Mount,
    slot: str,
    module: ModuleType,
    module_id: str,
    instrument_id: Optional[str] = None,
    cal_status: Optional[
        Union[local_types.CalibrationStatus, v1.CalibrationStatus]
    ] = None,
) -> None:
    module_dir = config.get_opentrons_path("module_calibration_dir")
    if isinstance(cal_status, local_types.CalibrationStatus):
        cal_status_model = v1.CalibrationStatus(**asdict(cal_status))
    elif isinstance(cal_status, v1.CalibrationStatus):
        cal_status_model = cal_status
    else:
        cal_status_model = v1.CalibrationStatus()

    module_calibration = v1.ModuleOffsetModel(
        offset=offset,
        mount=mount,
        slot=slot,
        module=module,
        module_id=module_id,
        instrument_id=instrument_id,
        lastModified=utc_now(),
        source=local_types.SourceType.user,
        status=cal_status_model,
    )
    io.save_to_file(module_dir, module_id, module_calibration)
    # something changed, clear lru cache
    get_module_offset.cache_clear()
    load_all_module_offsets.cache_clear()


# Get Module Offset Calibrations


@no_type_check
@lru_cache(maxsize=10)
def get_module_offset(
    module: ModuleType, module_id: str, slot: Optional[str] = None
) -> Optional[v1.ModuleOffsetModel]:
    try:
        module_calibration_filepath = (
            config.get_opentrons_path("module_calibration_dir") / f"{module_id}.json"
        )
        return v1.ModuleOffsetModel(**io.read_cal_file(module_calibration_filepath))
    except FileNotFoundError:
        log.warning(
            f"Calibrations for {module} {module_id} on slot {slot} does not exist."
        )
        return None
    except (json.JSONDecodeError, ValidationError):
        log.warning(
            f"Malformed calibrations for {module_id} on slot {slot}. Please factory reset your calibrations.",
            exc_info=True,
        )
        return None


@lru_cache(maxsize=10)
def load_all_module_offsets() -> List[v1.ModuleOffsetModel]:
    """Load all module offsets from the disk."""

    calibrations: List[v1.ModuleOffsetModel] = []
    files = os.listdir(config.get_opentrons_path("module_calibration_dir"))
    for file in files:
        try:
            calibrations.append(
                v1.ModuleOffsetModel(
                    **io.read_cal_file(
                        Path(config.get_opentrons_path("module_calibration_dir") / file)
                    )
                )
            )
        except (json.JSONDecodeError, ValidationError):
            log.warning(
                f"Malformed module calibrations for {file}. Please factory reset your calibrations.",
                exc_info=True,
            )
            continue
    return calibrations



"""Tip state tracking."""
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional, List

from .abstract_store import HasState, HandlesActions
from ..actions import (
    Action,
    UpdateCommandAction,
    ResetTipsAction,
)
from ..commands import (
    Command,
    LoadLabwareResult,
    PickUpTipResult,
    DropTipResult,
    DropTipInPlaceResult,
)
from ..commands.configuring_common import (
    PipetteConfigUpdateResultMixin,
    PipetteNozzleLayoutResultMixin,
)


class TipRackWellState(Enum):
    """The state of a single tip in a tip rack's well."""

    CLEAN = "clean"
    USED = "used"


TipRackStateByWellName = Dict[str, TipRackWellState]


@dataclass
class TipState:
    """State of all tips."""

    tips_by_labware_id: Dict[str, TipRackStateByWellName]
    column_by_labware_id: Dict[str, List[List[str]]]
    channels_by_pipette_id: Dict[str, int]
    length_by_pipette_id: Dict[str, float]
    active_channels_by_pipette_id: Dict[str, int]


class TipStore(HasState[TipState], HandlesActions):
    """Tip state container."""

    _state: TipState

    def __init__(self) -> None:
        """Initialize a liquid store and its state."""
        self._state = TipState(
            tips_by_labware_id={},
            column_by_labware_id={},
            channels_by_pipette_id={},
            length_by_pipette_id={},
            active_channels_by_pipette_id={},
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, UpdateCommandAction):
            if isinstance(action.private_result, PipetteConfigUpdateResultMixin):
                pipette_id = action.private_result.pipette_id
                config = action.private_result.config
                self._state.channels_by_pipette_id[pipette_id] = config.channels
                self._state.active_channels_by_pipette_id[pipette_id] = config.channels
            self._handle_command(action.command)

            if isinstance(action.private_result, PipetteNozzleLayoutResultMixin):
                pipette_id = action.private_result.pipette_id
                nozzle_map = action.private_result.nozzle_map
                if nozzle_map:
                    self._state.active_channels_by_pipette_id[
                        pipette_id
                    ] = nozzle_map.tip_count
                else:
                    self._state.active_channels_by_pipette_id[
                        pipette_id
                    ] = self._state.channels_by_pipette_id[pipette_id]

        elif isinstance(action, ResetTipsAction):
            labware_id = action.labware_id

            for well_name in self._state.tips_by_labware_id[labware_id].keys():
                self._state.tips_by_labware_id[labware_id][
                    well_name
                ] = TipRackWellState.CLEAN

    def _handle_command(self, command: Command) -> None:
        if (
            isinstance(command.result, LoadLabwareResult)
            and command.result.definition.parameters.isTiprack
        ):
            labware_id = command.result.labwareId
            definition = command.result.definition
            self._state.tips_by_labware_id[labware_id] = {
                well_name: TipRackWellState.CLEAN
                for column in definition.ordering
                for well_name in column
            }
            self._state.column_by_labware_id[labware_id] = [
                column for column in definition.ordering
            ]

        elif isinstance(command.result, PickUpTipResult):
            labware_id = command.params.labwareId
            well_name = command.params.wellName
            pipette_id = command.params.pipetteId
            length = command.result.tipLength
            self._set_used_tips(
                pipette_id=pipette_id, well_name=well_name, labware_id=labware_id
            )
            self._state.length_by_pipette_id[pipette_id] = length

        elif isinstance(command.result, (DropTipResult, DropTipInPlaceResult)):
            pipette_id = command.params.pipetteId
            self._state.length_by_pipette_id.pop(pipette_id, None)

    def _set_used_tips(self, pipette_id: str, well_name: str, labware_id: str) -> None:
        pipette_channels = self._state.active_channels_by_pipette_id.get(pipette_id)
        columns = self._state.column_by_labware_id.get(labware_id, [])
        wells = self._state.tips_by_labware_id.get(labware_id, {})

        if pipette_channels == len(wells):
            for well_name in wells.keys():
                wells[well_name] = TipRackWellState.USED

        elif columns and pipette_channels == len(columns[0]):
            for column in columns:
                if well_name in column:
                    for well in column:
                        wells[well] = TipRackWellState.USED
                    break

        else:
            wells[well_name] = TipRackWellState.USED


class TipView(HasState[TipState]):
    """Read-only tip state view."""

    _state: TipState

    def __init__(self, state: TipState) -> None:
        """Initialize the computed view of liquid state.

        Arguments:
            state: Liquid state dataclass used for all calculations.
        """
        self._state = state

    # TODO (spp, 2023-12-05): update this logic once we support partial nozzle configurations
    #  that require the tip tracking to move right to left or front to back;
    #  for example when using leftmost column config of 96-channel
    #  or backmost single nozzle configuration of an 8-channel.
    def get_next_tip(  # noqa: C901
        self, labware_id: str, num_tips: int, starting_tip_name: Optional[str]
    ) -> Optional[str]:
        """Get the next available clean tip."""
        wells = self._state.tips_by_labware_id.get(labware_id, {})
        columns = self._state.column_by_labware_id.get(labware_id, [])

        if columns and num_tips == len(columns[0]):  # Get next tips for 8-channel
            column_head = [column[0] for column in columns]
            starting_column_index = 0

            if starting_tip_name:
                for idx, column in enumerate(columns):
                    if starting_tip_name in column:
                        if starting_tip_name not in column_head:
                            starting_column_index = idx + 1
                        else:
                            starting_column_index = idx

            for column in columns[starting_column_index:]:
                if not any(wells[well] == TipRackWellState.USED for well in column):
                    return column[0]

        elif num_tips == len(wells.keys()):  # Get next tips for 96 channel
            if starting_tip_name and starting_tip_name != columns[0][0]:
                return None

            if not any(
                tip_state == TipRackWellState.USED for tip_state in wells.values()
            ):
                return next(iter(wells))

        else:  # Get next tips for single channel
            if starting_tip_name is not None:
                wells = _drop_wells_before_starting_tip(wells, starting_tip_name)

            for well_name, tip_state in wells.items():
                if tip_state == TipRackWellState.CLEAN:
                    return well_name

        return None

    def get_pipette_channels(self, pipette_id: str) -> int:
        """Return the given pipette's number of channels."""
        return self._state.channels_by_pipette_id[pipette_id]

    def get_pipette_active_channels(self, pipette_id: str) -> int:
        """Get the number of channels being used in the given pipette's configuration."""
        return self._state.active_channels_by_pipette_id[pipette_id]

    def has_clean_tip(self, labware_id: str, well_name: str) -> bool:
        """Get whether a well in a labware has a clean tip.

        Args:
            labware_id: The labware ID to check.
            well_name: The well name to check.

        Returns:
            True if the labware is a tip rack and the well has a clean tip,
            otherwise False.
        """
        tip_rack = self._state.tips_by_labware_id.get(labware_id)
        well_state = tip_rack.get(well_name) if tip_rack else None

        return well_state == TipRackWellState.CLEAN

    def get_tip_length(self, pipette_id: str) -> float:
        """Return the given pipette's tip length."""
        return self._state.length_by_pipette_id.get(pipette_id, 0)


def _drop_wells_before_starting_tip(
    wells: TipRackStateByWellName, starting_tip_name: str
) -> TipRackStateByWellName:
    """Drop any wells that come before the starting tip and return the remaining ones after."""
    seen_starting_well = False
    remaining_wells = {}
    for well_name, tip_state in wells.items():
        if well_name == starting_tip_name:
            seen_starting_well = True
        if seen_starting_well:
            remaining_wells[well_name] = tip_state
    return remaining_wells



"""Deck state accessors for the Protocol API."""
from dataclasses import dataclass
from typing import Iterator, List, Mapping, Optional, Tuple, Union

from opentrons_shared_data.deck.dev_types import SlotDefV3

from opentrons.motion_planning import adjacent_slots_getters
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import APIVersionError
from opentrons.types import DeckLocation, DeckSlotName, StagingSlotName, Location, Point
from opentrons_shared_data.robot.dev_types import RobotType


from .core.common import ProtocolCore
from .core.core_map import LoadedCoreMap
from .core.module import AbstractModuleCore
from .labware import Labware
from .module_contexts import ModuleContext
from ._types import OFF_DECK
from . import validation


DeckItem = Union[Labware, ModuleContext]

STAGING_SLOT_VERSION_GATE = APIVersion(2, 16)


@dataclass(frozen=True)
class CalibrationPosition:
    """A calibration point on the deck of the robot.

    Attributes:
        id: Unique identifier for the calibration point.
        position: The absolute x, y, z coordinate of the point.
        displayName: A human-readable nickname for this point.
    """

    id: str
    position: Tuple[float, float, float]
    displayName: str


def _get_slot_name(
    slot_key: DeckLocation, api_version: APIVersion, robot_type: RobotType
) -> Union[DeckSlotName, StagingSlotName]:
    try:
        slot = validation.ensure_and_convert_deck_slot(
            slot_key, api_version, robot_type
        )
        return slot
    except (TypeError, ValueError) as error:
        raise KeyError(slot_key) from error


class Deck(Mapping[DeckLocation, Optional[DeckItem]]):
    """A dictionary-like object to access Protocol API objects loaded on the deck.

    Accessible via :py:meth:`ProtocolContext.deck`.
    """

    def __init__(
        self,
        protocol_core: ProtocolCore,
        core_map: LoadedCoreMap,
        api_version: APIVersion,
    ) -> None:
        self._protocol_core = protocol_core
        self._core_map = core_map
        self._api_version = api_version

        deck_locations = protocol_core.get_deck_definition()["locations"]

        self._slot_definitions_by_name = self._protocol_core.get_slot_definitions()
        if self._api_version >= STAGING_SLOT_VERSION_GATE:
            self._slot_definitions_by_name.update(
                self._protocol_core.get_staging_slot_definitions()
            )

        self._calibration_positions = [
            CalibrationPosition(
                id=point["id"],
                displayName=point["displayName"],
                position=(
                    point["position"][0],
                    point["position"][1],
                    point["position"][2],
                ),
            )
            for point in deck_locations["calibrationPoints"]
        ]

    def __getitem__(self, key: DeckLocation) -> Optional[DeckItem]:
        """Get the item, if any, located in a given slot."""
        slot_name = _get_slot_name(
            key, self._api_version, self._protocol_core.robot_type
        )
        item_core = self._protocol_core.get_slot_item(slot_name)
        item = self._core_map.get(item_core)

        return item

    def __delitem__(self, key: DeckLocation) -> None:
        if self._api_version == APIVersion(2, 14):
            # __delitem__() support history:
            #
            # * PAPIv<=2.13 (non Protocol Engine): Yes, but that goes through a different Deck class
            # * PAPIv2.14 (Protocol Engine): No
            # * PAPIv2.15 (Protocol Engine): Yes
            raise APIVersionError(
                f"Deleting deck elements is not supported with apiLevel {self._api_version}."
                f" Try increasing your apiLevel to {APIVersion(2, 15)}."
            )

        slot_name = _get_slot_name(
            key, self._api_version, self._protocol_core.robot_type
        )
        item_core = self._protocol_core.get_slot_item(slot_name)

        if item_core is None:
            # No-op if trying to delete from an empty slot.
            # This matches pre-Protocol-Engine (PAPIv<=2.13) behavior.
            pass
        elif isinstance(item_core, AbstractModuleCore):
            # Protocol Engine does not support removing modules from the deck.
            # This is a change from pre-Protocol-Engine (PAPIv<=2.13) behavior, unfortunately.
            raise TypeError(
                f"Slot {repr(key)} contains a module, {item_core.get_display_name()}."
                f" You can only delete labware, not modules."
            )
        else:
            self._protocol_core.move_labware(
                item_core,
                new_location=OFF_DECK,
                use_gripper=False,
                pause_for_manual_move=False,
                pick_up_offset=None,
                drop_offset=None,
            )

    def __iter__(self) -> Iterator[str]:
        """Iterate through all deck slots."""
        return iter(self._slot_definitions_by_name)

    def __len__(self) -> int:
        """Get the number of slots on the deck."""
        return len(self._slot_definitions_by_name)

    # todo(mm, 2023-05-08): This may be internal and removable from this public class. Jira RSS-236.
    def right_of(self, slot: DeckLocation) -> Optional[DeckItem]:
        """Get the item directly to the right of the given slot, if any."""
        slot_name = _get_slot_name(
            slot, self._api_version, self._protocol_core.robot_type
        )
        if isinstance(slot_name, DeckSlotName):
            east_slot = adjacent_slots_getters.get_east_slot(slot_name.as_int())
        else:
            east_slot = None

        return self[east_slot] if east_slot is not None else None

    # todo(mm, 2023-05-08): This may be internal and removable from this public class. Jira RSS-236.
    def left_of(self, slot: DeckLocation) -> Optional[DeckItem]:
        """Get the item directly to the left of the given slot, if any."""
        slot_name = _get_slot_name(
            slot, self._api_version, self._protocol_core.robot_type
        )
        west_slot: Optional[DeckLocation]
        if isinstance(slot_name, DeckSlotName):
            west_slot = adjacent_slots_getters.get_west_slot(slot_name.as_int())
        else:
            west_slot = adjacent_slots_getters.get_west_of_staging_slot(slot_name).id

        return self[west_slot] if west_slot is not None else None

    # todo(mm, 2023-05-08): This is undocumented in the public PAPI, but is used in some protocols
    # written by Applications Engineering. Either officially document this, or decide it's internal
    # and remove it from this class. Jira RSS-236.
    def position_for(self, slot: DeckLocation) -> Location:
        """Get the absolute location of a deck slot's front-left corner."""
        slot_name = _get_slot_name(
            slot, self._api_version, self._protocol_core.robot_type
        )
        slot_definition = self._slot_definitions_by_name[slot_name.id]
        x, y, z = slot_definition["position"]
        normalized_slot_name = validation.internal_slot_to_public_string(
            slot_name, self._protocol_core.robot_type
        )
        return Location(point=Point(x, y, z), labware=normalized_slot_name)

    # todo(mm, 2023-05-08): This may be internal and removable from this public class. Jira RSS-236.
    def get_slot_definition(self, slot: DeckLocation) -> SlotDefV3:
        """Get the geometric definition data of a slot."""
        slot_name = _get_slot_name(
            slot, self._api_version, self._protocol_core.robot_type
        )
        return self._slot_definitions_by_name[slot_name.id]

    # todo(mm, 2023-05-08): This may be internal and removable from this public class. Jira RSS-236.
    def get_slot_center(self, slot: DeckLocation) -> Point:
        """Get the absolute coordinates of a slot's center."""
        slot_name = _get_slot_name(
            slot, self._api_version, self._protocol_core.robot_type
        )
        return self._protocol_core.get_slot_center(slot_name)

    # todo(mm, 2023-05-08): This may be internal and removable from this public class. Jira RSS-236.
    @property
    def highest_z(self) -> float:
        """Get the height of the tallest known point on the deck."""
        return self._protocol_core.get_highest_z()

    # todo(mm, 2023-05-08): This appears internal. Remove it from this public class. Jira RSS-236.
    @property
    def slots(self) -> List[SlotDefV3]:
        """Get a list of all slot definitions."""
        return list(self._slot_definitions_by_name.values())

    # todo(mm, 2023-05-08): This appears internal. Remove it from this public class. Jira RSS-236.
    @property
    def calibration_positions(self) -> List[CalibrationPosition]:
        """Get a list of all calibration positions on the deck."""
        return list(self._calibration_positions)



"""Basic pipette data state and store."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Dict, List, Mapping, Optional, Tuple

from opentrons_shared_data.pipette import pipette_definition
from opentrons.config.defaults_ot2 import Z_RETRACT_DISTANCE
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.nozzle_manager import (
    NozzleConfigurationType,
    NozzleMap,
)
from opentrons.types import MountType, Mount as HwMount, Point

from .. import errors
from ..types import (
    LoadedPipette,
    MotorAxis,
    FlowRates,
    DeckPoint,
    CurrentWell,
    CurrentAddressableArea,
    CurrentPipetteLocation,
    TipGeometry,
)
from ..commands import (
    Command,
    LoadPipetteResult,
    AspirateResult,
    AspirateInPlaceResult,
    DispenseResult,
    DispenseInPlaceResult,
    MoveLabwareResult,
    MoveToCoordinatesResult,
    MoveToWellResult,
    MoveRelativeResult,
    MoveToAddressableAreaResult,
    MoveToAddressableAreaForDropTipResult,
    PickUpTipResult,
    DropTipResult,
    DropTipInPlaceResult,
    HomeResult,
    RetractAxisResult,
    BlowOutResult,
    BlowOutInPlaceResult,
    TouchTipResult,
    thermocycler,
    heater_shaker,
    CommandPrivateResult,
    PrepareToAspirateResult,
)
from ..commands.configuring_common import (
    PipetteConfigUpdateResultMixin,
    PipetteNozzleLayoutResultMixin,
)
from ..actions import (
    Action,
    SetPipetteMovementSpeedAction,
    UpdateCommandAction,
)
from .abstract_store import HasState, HandlesActions


@dataclass(frozen=True)
class HardwarePipette:
    """Hardware pipette data."""

    mount: HwMount
    config: PipetteDict


@dataclass(frozen=True)
class CurrentDeckPoint:
    """The latest deck point and mount the robot has accessed."""

    mount: Optional[MountType]
    deck_point: Optional[DeckPoint]


@dataclass(frozen=True)
class BoundingNozzlesOffsets:
    """Offsets of the bounding nozzles of the pipette."""

    back_left_offset: Point
    front_right_offset: Point


@dataclass(frozen=True)
class PipetteBoundingBoxOffsets:
    """Offsets of the corners of the pipette's bounding box."""

    back_left_corner: Point
    front_right_corner: Point


@dataclass(frozen=True)
class StaticPipetteConfig:
    """Static config for a pipette."""

    model: str
    serial_number: str
    display_name: str
    min_volume: float
    max_volume: float
    channels: int
    tip_configuration_lookup_table: Dict[
        float, pipette_definition.SupportedTipsDefinition
    ]
    nominal_tip_overlap: Dict[str, float]
    home_position: float
    nozzle_offset_z: float
    pipette_bounding_box_offsets: PipetteBoundingBoxOffsets
    bounding_nozzle_offsets: BoundingNozzlesOffsets


@dataclass
class PipetteState:
    """Basic pipette data state and getter methods."""

    pipettes_by_id: Dict[str, LoadedPipette]
    aspirated_volume_by_id: Dict[str, Optional[float]]
    current_location: Optional[CurrentPipetteLocation]
    current_deck_point: CurrentDeckPoint
    attached_tip_by_id: Dict[str, Optional[TipGeometry]]
    movement_speed_by_id: Dict[str, Optional[float]]
    static_config_by_id: Dict[str, StaticPipetteConfig]
    flow_rates_by_id: Dict[str, FlowRates]
    nozzle_configuration_by_id: Dict[str, Optional[NozzleMap]]


class PipetteStore(HasState[PipetteState], HandlesActions):
    """Pipette state container."""

    _state: PipetteState

    def __init__(self) -> None:
        """Initialize a PipetteStore and its state."""
        self._state = PipetteState(
            pipettes_by_id={},
            aspirated_volume_by_id={},
            attached_tip_by_id={},
            current_location=None,
            current_deck_point=CurrentDeckPoint(mount=None, deck_point=None),
            movement_speed_by_id={},
            static_config_by_id={},
            flow_rates_by_id={},
            nozzle_configuration_by_id={},
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, UpdateCommandAction):
            self._handle_command(action.command, action.private_result)
        elif isinstance(action, SetPipetteMovementSpeedAction):
            self._state.movement_speed_by_id[action.pipette_id] = action.speed

    def _handle_command(  # noqa: C901
        self, command: Command, private_result: CommandPrivateResult
    ) -> None:
        self._update_current_location(command)
        self._update_deck_point(command)

        if isinstance(private_result, PipetteConfigUpdateResultMixin):
            config = private_result.config
            self._state.static_config_by_id[
                private_result.pipette_id
            ] = StaticPipetteConfig(
                serial_number=private_result.serial_number,
                model=config.model,
                display_name=config.display_name,
                min_volume=config.min_volume,
                max_volume=config.max_volume,
                channels=config.channels,
                tip_configuration_lookup_table=config.tip_configuration_lookup_table,
                nominal_tip_overlap=config.nominal_tip_overlap,
                home_position=config.home_position,
                nozzle_offset_z=config.nozzle_offset_z,
                pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                    back_left_corner=config.back_left_corner_offset,
                    front_right_corner=config.front_right_corner_offset,
                ),
                bounding_nozzle_offsets=BoundingNozzlesOffsets(
                    back_left_offset=config.back_left_nozzle_offset,
                    front_right_offset=config.front_right_nozzle_offset,
                ),
            )
            self._state.flow_rates_by_id[private_result.pipette_id] = config.flow_rates
        elif isinstance(private_result, PipetteNozzleLayoutResultMixin):
            self._state.nozzle_configuration_by_id[
                private_result.pipette_id
            ] = private_result.nozzle_map

        if isinstance(command.result, LoadPipetteResult):
            pipette_id = command.result.pipetteId

            self._state.pipettes_by_id[pipette_id] = LoadedPipette(
                id=pipette_id,
                pipetteName=command.params.pipetteName,
                mount=command.params.mount,
            )
            self._state.aspirated_volume_by_id[pipette_id] = None
            self._state.movement_speed_by_id[pipette_id] = None
            self._state.attached_tip_by_id[pipette_id] = None
            self._state.nozzle_configuration_by_id[pipette_id] = None

        elif isinstance(command.result, (AspirateResult, AspirateInPlaceResult)):
            pipette_id = command.params.pipetteId
            previous_volume = self._state.aspirated_volume_by_id[pipette_id] or 0
            # PipetteHandler will have clamped command.result.volume for us, so
            # next_volume should always be in bounds.
            next_volume = previous_volume + command.result.volume

            self._state.aspirated_volume_by_id[pipette_id] = next_volume

        elif isinstance(command.result, (DispenseResult, DispenseInPlaceResult)):
            pipette_id = command.params.pipetteId
            previous_volume = self._state.aspirated_volume_by_id[pipette_id] or 0
            # PipetteHandler will have clamped command.result.volume for us, so
            # next_volume should always be in bounds.
            next_volume = previous_volume - command.result.volume
            self._state.aspirated_volume_by_id[pipette_id] = next_volume

        elif isinstance(command.result, PickUpTipResult):
            pipette_id = command.params.pipetteId
            attached_tip = TipGeometry(
                length=command.result.tipLength,
                volume=command.result.tipVolume,
                diameter=command.result.tipDiameter,
            )

            self._state.attached_tip_by_id[pipette_id] = attached_tip
            self._state.aspirated_volume_by_id[pipette_id] = 0

            static_config = self._state.static_config_by_id.get(pipette_id)
            if static_config:
                try:
                    tip_configuration = static_config.tip_configuration_lookup_table[
                        attached_tip.volume
                    ]
                except KeyError:
                    # TODO(seth,9/11/2023): this is a bad way of doing defaults but better than max volume.
                    # we used to look up a default tip config via the pipette max volume, but if that isn't
                    # tip volume (as it isn't when we're in low-volume mode) then that lookup fails. Using
                    # the first entry in the table is ok I guess but we really need to generally rethink how
                    # we identify tip classes - looking things up by volume is not enough.
                    tip_configuration = list(
                        static_config.tip_configuration_lookup_table.values()
                    )[0]
                self._state.flow_rates_by_id[pipette_id] = FlowRates(
                    default_blow_out=tip_configuration.default_blowout_flowrate.values_by_api_level,
                    default_aspirate=tip_configuration.default_aspirate_flowrate.values_by_api_level,
                    default_dispense=tip_configuration.default_dispense_flowrate.values_by_api_level,
                )

        elif isinstance(command.result, (DropTipResult, DropTipInPlaceResult)):
            pipette_id = command.params.pipetteId
            self._state.aspirated_volume_by_id[pipette_id] = None
            self._state.attached_tip_by_id[pipette_id] = None

            static_config = self._state.static_config_by_id.get(pipette_id)
            if static_config:
                # TODO(seth,9/11/2023): bad way to do defaulting, see above.
                tip_configuration = list(
                    static_config.tip_configuration_lookup_table.values()
                )[0]
                self._state.flow_rates_by_id[pipette_id] = FlowRates(
                    default_blow_out=tip_configuration.default_blowout_flowrate.values_by_api_level,
                    default_aspirate=tip_configuration.default_aspirate_flowrate.values_by_api_level,
                    default_dispense=tip_configuration.default_dispense_flowrate.values_by_api_level,
                )
        elif isinstance(command.result, (BlowOutResult, BlowOutInPlaceResult)):
            pipette_id = command.params.pipetteId
            self._state.aspirated_volume_by_id[pipette_id] = None

        elif isinstance(command.result, PrepareToAspirateResult):
            pipette_id = command.params.pipetteId
            self._state.aspirated_volume_by_id[pipette_id] = 0

    def _update_current_location(self, command: Command) -> None:
        # These commands leave the pipette in a new location.
        # Update current_location to reflect that.
        if isinstance(
            command.result,
            (
                MoveToWellResult,
                PickUpTipResult,
                DropTipResult,
                AspirateResult,
                DispenseResult,
                BlowOutResult,
                TouchTipResult,
            ),
        ):
            self._state.current_location = CurrentWell(
                pipette_id=command.params.pipetteId,
                labware_id=command.params.labwareId,
                well_name=command.params.wellName,
            )

        elif isinstance(
            command.result,
            (MoveToAddressableAreaResult, MoveToAddressableAreaForDropTipResult),
        ):
            self._state.current_location = CurrentAddressableArea(
                pipette_id=command.params.pipetteId,
                addressable_area_name=command.params.addressableAreaName,
            )

        # These commands leave the pipette in a place that we can't logically associate
        # with a well. Clear current_location to reflect the fact that it's now unknown.
        #
        # TODO(mc, 2021-11-12): Wipe out current_location on movement failures, too.
        # TODO(jbl 2023-02-14): Need to investigate whether move relative should clear current location
        elif isinstance(
            command.result,
            (
                HomeResult,
                RetractAxisResult,
                MoveToCoordinatesResult,
                thermocycler.OpenLidResult,
                thermocycler.CloseLidResult,
            ),
        ):
            self._state.current_location = None

        # Heater-Shaker commands may have left the pipette in a place that we can't
        # associate with a logical location, depending on their result.
        elif isinstance(
            command.result,
            (
                heater_shaker.SetAndWaitForShakeSpeedResult,
                heater_shaker.OpenLabwareLatchResult,
            ),
        ):
            if command.result.pipetteRetracted:
                self._state.current_location = None

        # A moveLabware command may have moved the labware that contains the current
        # well out from under the pipette. Clear the current location to reflect the
        # fact that the pipette is no longer over any labware.
        #
        # This is necessary for safe motion planning in case the next movement
        # goes to the same labware (now in a new place).
        elif isinstance(command.result, MoveLabwareResult):
            moved_labware_id = command.params.labwareId
            if command.params.strategy == "usingGripper":
                # All mounts will have been retracted.
                self._state.current_location = None
            elif (
                isinstance(self._state.current_location, CurrentWell)
                and self._state.current_location.labware_id == moved_labware_id
            ):
                self._state.current_location = None

    def _update_deck_point(self, command: Command) -> None:
        if isinstance(
            command.result,
            (
                MoveToWellResult,
                MoveToCoordinatesResult,
                MoveRelativeResult,
                MoveToAddressableAreaResult,
                MoveToAddressableAreaForDropTipResult,
                PickUpTipResult,
                DropTipResult,
                AspirateResult,
                DispenseResult,
                BlowOutResult,
                TouchTipResult,
            ),
        ):
            pipette_id = command.params.pipetteId
            deck_point = command.result.position

            try:
                loaded_pipette = self._state.pipettes_by_id[pipette_id]
            except KeyError:
                self._clear_deck_point()
            else:
                self._state.current_deck_point = CurrentDeckPoint(
                    mount=loaded_pipette.mount, deck_point=deck_point
                )

        elif isinstance(
            command.result,
            (
                HomeResult,
                RetractAxisResult,
                thermocycler.OpenLidResult,
                thermocycler.CloseLidResult,
            ),
        ):
            self._clear_deck_point()

        elif isinstance(
            command.result,
            (
                heater_shaker.SetAndWaitForShakeSpeedResult,
                heater_shaker.OpenLabwareLatchResult,
            ),
        ):
            if command.result.pipetteRetracted:
                self._clear_deck_point()

        elif isinstance(command.result, MoveLabwareResult):
            if command.params.strategy == "usingGripper":
                # All mounts will have been retracted.
                self._clear_deck_point()

    def _clear_deck_point(self) -> None:
        """Reset last deck point to default None value for mount and point."""
        self._state.current_deck_point = CurrentDeckPoint(mount=None, deck_point=None)


class PipetteView(HasState[PipetteState]):
    """Read-only view of computed pipettes state."""

    _state: PipetteState

    def __init__(self, state: PipetteState) -> None:
        """Initialize the view with its backing state value."""
        self._state = state

    def get(self, pipette_id: str) -> LoadedPipette:
        """Get pipette data by the pipette's unique identifier."""
        try:
            return self._state.pipettes_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found."
            ) from e

    def get_mount(self, pipette_id: str) -> MountType:
        """Get the pipette's mount."""
        return self.get(pipette_id).mount

    def get_all(self) -> List[LoadedPipette]:
        """Get a list of all pipette entries in state."""
        return list(self._state.pipettes_by_id.values())

    def get_by_mount(self, mount: MountType) -> Optional[LoadedPipette]:
        """Get pipette data by the pipette's mount."""
        for pipette in self._state.pipettes_by_id.values():
            if pipette.mount == mount:
                return pipette
        return None

    def get_hardware_pipette(
        self,
        pipette_id: str,
        attached_pipettes: Mapping[HwMount, Optional[PipetteDict]],
    ) -> HardwarePipette:
        """Get a pipette's hardware configuration and state by ID."""
        pipette_data = self.get(pipette_id)
        pipette_name = pipette_data.pipetteName
        mount = pipette_data.mount

        hw_mount = mount.to_hw_mount()
        hw_config = attached_pipettes[hw_mount]

        # TODO(mc, 2022-01-11): HW controller may return an empty dict for
        # no pipette attached instead of `None`. Update when fixed in HWAPI
        if not hw_config:
            raise errors.PipetteNotAttachedError(f"No pipette attached on {mount}")

        elif (
            hw_config["name"] != pipette_name
            and pipette_name not in hw_config["back_compat_names"]
        ):
            raise errors.PipetteNotAttachedError(
                f"Found {hw_config['name']} on {mount}, "
                f"but {pipette_id} is a {pipette_name}"
            )

        return HardwarePipette(mount=hw_mount, config=hw_config)

    def get_current_location(self) -> Optional[CurrentPipetteLocation]:
        """Get the last accessed location and which pipette accessed it."""
        return self._state.current_location

    def get_deck_point(self, pipette_id: str) -> Optional[DeckPoint]:
        """Get the deck point of a pipette by ID, or None if it was not associated with the last move operation."""
        loaded_pipette = self.get(pipette_id)
        current_deck_point = self._state.current_deck_point
        if loaded_pipette.mount == current_deck_point.mount:
            return current_deck_point.deck_point
        return None

    def get_attached_tip(self, pipette_id: str) -> Optional[TipGeometry]:
        """Get details of the pipette's attached tip.

        Returns:
            The tip's volume and length, or None if there is no tip attached,
        """
        try:
            return self._state.attached_tip_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} no found; unable to get attached tip."
            ) from e

    def get_all_attached_tips(self) -> List[Tuple[str, TipGeometry]]:
        """Get a list of all attached tips.

        Returns:
            A list of pipette ID, tip details tuples.
        """
        return [
            (pipette_id, tip)
            for pipette_id, tip in self._state.attached_tip_by_id.items()
            if tip is not None
        ]

    def get_aspirated_volume(self, pipette_id: str) -> Optional[float]:
        """Get the currently aspirated volume of a pipette by ID.

        Returns:
            The volume the pipette has aspirated.
            None, after blow-out and the plunger is in an unsafe position or drop-tip and there is no tip attached.

        Raises:
            PipetteNotLoadedError: pipette ID does not exist.
            TipNotAttachedError: if no tip is attached to the pipette.
        """
        self.validate_tip_state(pipette_id, True)

        try:
            return self._state.aspirated_volume_by_id[pipette_id]

        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get current volume."
            ) from e

    def get_working_volume(self, pipette_id: str) -> float:
        """Get the working maximum volume of a pipette by ID.

        Raises:
            PipetteNotLoadedError: pipette ID does not exist.
            TipNotAttachedError: if no tip is attached to the pipette.
        """
        max_volume = self.get_maximum_volume(pipette_id)
        attached_tip = self.get_attached_tip(pipette_id)

        if not attached_tip:
            raise errors.TipNotAttachedError(
                f"Pipette {pipette_id} has no tip attached; unable to calculate working maximum volume."
            )

        return min(attached_tip.volume, max_volume)

    def get_available_volume(self, pipette_id: str) -> Optional[float]:
        """Get the available volume of a pipette by ID."""
        working_volume = self.get_working_volume(pipette_id)
        current_volume = self.get_aspirated_volume(pipette_id)

        return max(0.0, working_volume - current_volume) if current_volume else None

    def validate_tip_state(self, pipette_id: str, expected_has_tip: bool) -> None:
        """Validate that a pipette's tip state matches expectations."""
        attached_tip = self.get_attached_tip(pipette_id)

        if expected_has_tip is True and attached_tip is None:
            raise errors.TipNotAttachedError(
                "Pipette should have a tip attached, but does not."
            )
        if expected_has_tip is False and attached_tip is not None:
            raise errors.TipAttachedError(
                "Pipette should not have a tip attached, but does."
            )

    def get_movement_speed(
        self, pipette_id: str, requested_speed: Optional[float] = None
    ) -> Optional[float]:
        """Return the given pipette's requested or current movement speed."""
        return requested_speed or self._state.movement_speed_by_id[pipette_id]

    def get_config(self, pipette_id: str) -> StaticPipetteConfig:
        """Get the static pipette configuration by pipette id."""
        try:
            return self._state.static_config_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get pipette configuration."
            ) from e

    def get_model_name(self, pipette_id: str) -> str:
        """Return the given pipette's model name."""
        return self.get_config(pipette_id).model

    def get_display_name(self, pipette_id: str) -> str:
        """Return the given pipette's display name."""
        return self.get_config(pipette_id).display_name

    def get_serial_number(self, pipette_id: str) -> str:
        """Get the serial number of the pipette."""
        return self.get_config(pipette_id).serial_number

    def get_channels(self, pipette_id: str) -> int:
        """Return the max channels of the pipette."""
        return self.get_config(pipette_id).channels

    def get_minimum_volume(self, pipette_id: str) -> float:
        """Return the given pipette's minimum volume."""
        return self.get_config(pipette_id).min_volume

    def get_maximum_volume(self, pipette_id: str) -> float:
        """Return the given pipette's maximum volume."""
        return self.get_config(pipette_id).max_volume

    def get_instrument_max_height_ot2(self, pipette_id: str) -> float:
        """Get calculated max instrument height for an OT-2."""
        config = self.get_config(pipette_id)
        return config.home_position - Z_RETRACT_DISTANCE + config.nozzle_offset_z

    def get_return_tip_scale(self, pipette_id: str) -> float:
        """Return the given pipette's return tip height scale."""
        max_volume = self.get_maximum_volume(pipette_id)
        working_volume = max_volume
        tip = self.get_attached_tip(pipette_id)
        if tip:
            working_volume = tip.volume

        if working_volume in self.get_config(pipette_id).tip_configuration_lookup_table:
            tip_lookup = self.get_config(pipette_id).tip_configuration_lookup_table[
                working_volume
            ]
        else:
            tip_lookup = self.get_config(pipette_id).tip_configuration_lookup_table[
                max_volume
            ]
        return tip_lookup.default_return_tip_height

    def get_flow_rates(self, pipette_id: str) -> FlowRates:
        """Get the default flow rates for the pipette."""
        try:
            return self._state.flow_rates_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get pipette flow rates."
            ) from e

    def get_nominal_tip_overlap(self, pipette_id: str, labware_uri: str) -> float:
        """Get the nominal tip overlap for a given labware from config."""
        tip_overlaps_by_uri = self.get_config(pipette_id).nominal_tip_overlap

        try:
            return tip_overlaps_by_uri[labware_uri]
        except KeyError:
            return tip_overlaps_by_uri.get("default", 0)

    def get_z_axis(self, pipette_id: str) -> MotorAxis:
        """Get the MotorAxis representing this pipette's Z stage."""
        mount = self.get(pipette_id).mount
        return MotorAxis.LEFT_Z if mount == MountType.LEFT else MotorAxis.RIGHT_Z

    def get_plunger_axis(self, pipette_id: str) -> MotorAxis:
        """Get the MotorAxis representing this pipette's plunger."""
        mount = self.get(pipette_id).mount
        return (
            MotorAxis.LEFT_PLUNGER
            if mount == MountType.LEFT
            else MotorAxis.RIGHT_PLUNGER
        )

    def get_nozzle_layout_type(self, pipette_id: str) -> NozzleConfigurationType:
        """Get the current set nozzle layout configuration."""
        nozzle_map_for_pipette = self._state.nozzle_configuration_by_id.get(pipette_id)
        if nozzle_map_for_pipette:
            return nozzle_map_for_pipette.configuration
        else:
            return NozzleConfigurationType.FULL

    def get_is_partially_configured(self, pipette_id: str) -> bool:
        """Determine if the provided pipette is partially configured."""
        return self.get_nozzle_layout_type(pipette_id) != NozzleConfigurationType.FULL

    def get_primary_nozzle(self, pipette_id: str) -> Optional[str]:
        """Get the primary nozzle, if any, related to the given pipette's nozzle configuration."""
        nozzle_map = self._state.nozzle_configuration_by_id.get(pipette_id)
        return nozzle_map.starting_nozzle if nozzle_map else None

    def get_primary_nozzle_offset(self, pipette_id: str) -> Point:
        """Get the pipette's current primary nozzle's offset."""
        nozzle_map = self._state.nozzle_configuration_by_id.get(pipette_id)
        if nozzle_map:
            primary_nozzle_offset = nozzle_map.starting_nozzle_offset
        else:
            # When not in partial configuration, back-left nozzle is the primary
            primary_nozzle_offset = self.get_config(
                pipette_id
            ).bounding_nozzle_offsets.back_left_offset
        return primary_nozzle_offset

    def get_pipette_bounding_nozzle_offsets(
        self, pipette_id: str
    ) -> BoundingNozzlesOffsets:
        """Get the nozzle offsets of the pipette's bounding nozzles."""
        return self.get_config(pipette_id).bounding_nozzle_offsets

    def get_pipette_bounds_at_specified_move_to_position(
        self,
        pipette_id: str,
        destination_position: Point,
    ) -> Tuple[Point, Point, Point, Point]:
        """Get the pipette's bounding offsets when primary nozzle is at the given position."""
        primary_nozzle_offset = self.get_primary_nozzle_offset(pipette_id)
        tip = self.get_attached_tip(pipette_id)
        # Primary nozzle position at destination, in deck coordinates
        primary_nozzle_position = destination_position + Point(
            x=0, y=0, z=tip.length if tip else 0
        )

        # Get the pipette bounding box based on total nozzles
        pipette_bounds_offsets = self.get_config(
            pipette_id
        ).pipette_bounding_box_offsets
        pip_back_left_bound = (
            primary_nozzle_position
            - primary_nozzle_offset
            + pipette_bounds_offsets.back_left_corner
        )
        pip_front_right_bound = (
            primary_nozzle_position
            - primary_nozzle_offset
            + pipette_bounds_offsets.front_right_corner
        )
        # TODO (spp, 2024-02-27): remove back right & front left;
        #  return only back left and front right points.
        pip_back_right_bound = Point(
            pip_front_right_bound.x, pip_back_left_bound.y, pip_front_right_bound.z
        )
        pip_front_left_bound = Point(
            pip_back_left_bound.x, pip_front_right_bound.y, pip_back_left_bound.z
        )
        return (
            pip_back_left_bound,
            pip_front_right_bound,
            pip_back_right_bound,
            pip_front_left_bound,
        )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional, Union

from opentrons import types
from opentrons.hardware_control import CriticalPoint
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocols.api_support import instrument as instrument_support
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import (
    build_edges,
    FlowRates,
    PlungerSpeeds,
    APIVersionError,
)
from opentrons.protocols.geometry import planning
from opentrons.protocol_api._nozzle_layout import NozzleLayout

from ..._trash_bin import TrashBin
from ..._waste_chute import WasteChute
from ..instrument import AbstractInstrument
from .legacy_well_core import LegacyWellCore
from .legacy_module_core import LegacyThermocyclerCore, LegacyHeaterShakerCore

if TYPE_CHECKING:
    from .legacy_protocol_core import LegacyProtocolCore


_log = logging.getLogger()

_PRE_2_2_TIP_DROP_HEIGHT_MM = 10
"""In PAPIv2.1 and below, tips are always dropped 10 mm from the bottom of the well."""


class LegacyInstrumentCore(AbstractInstrument[LegacyWellCore]):
    """Implementation of the InstrumentContext interface."""

    def __init__(
        self,
        protocol_interface: LegacyProtocolCore,
        mount: types.Mount,
        instrument_name: str,
        default_speed: float,
        api_version: Optional[APIVersion] = None,
    ):
        self._api_version = api_version or MAX_SUPPORTED_VERSION
        self._protocol_interface = protocol_interface
        self._mount = mount
        self._instrument_name = instrument_name
        self._default_speed = default_speed
        self._speeds = PlungerSpeeds(self)

        pipette_state = self.get_hardware_state()
        self._flow_rates = FlowRates(self)
        self._flow_rates.set_defaults(
            aspirate_defaults=pipette_state["default_aspirate_flow_rates"],
            dispense_defaults=pipette_state["default_dispense_flow_rates"],
            blow_out_defaults=pipette_state["default_blow_out_flow_rates"],
            api_level=self._api_version,
        )

    def get_default_speed(self) -> float:
        """Gets the speed at which the robot's gantry moves."""
        return self._default_speed

    def set_default_speed(self, speed: float) -> None:
        """Sets the speed at which the robot's gantry moves."""
        self._default_speed = speed

    def aspirate(
        self,
        location: types.Location,
        well_core: Optional[LegacyWellCore],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
    ) -> None:
        """Aspirate a given volume of liquid from the specified location.
        Args:
            volume: The volume of liquid to aspirate, in microliters.
            location: The exact location to aspirate from.
            well_core: The well to aspirate from, if applicable.
            rate: The rate in µL/s to aspirate at.
            flow_rate: Not used in this core.
            in_place: Whether we should move_to location.
        """
        if self.get_current_volume() == 0:
            # Make sure we're at the top of the labware and clear of any
            # liquid to prepare the pipette for aspiration
            if self._api_version < APIVersion(2, 3) or not self.is_ready_to_aspirate():
                if location.labware.is_well:
                    self.move_to(location=location.labware.as_well().top())
                else:
                    # TODO(seth,2019/7/29): This should be a warning exposed
                    #  via rpc to the runapp
                    _log.warning(
                        "When aspirate is called on something other than a "
                        "well relative position, we can't move to the top of"
                        " the well to prepare for aspiration. This might "
                        "cause over aspiration if the previous command is a "
                        "blow_out."
                    )
                self.prepare_to_aspirate()
            self.move_to(location=location)
        elif not in_place:
            self.move_to(location=location)

        self._protocol_interface.get_hardware().aspirate(self._mount, volume, rate)

    def dispense(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[LegacyWellCore],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
        push_out: Optional[float],
    ) -> None:
        """Dispense a given volume of liquid into the specified location.
        Args:
            volume: The volume of liquid to dispense, in microliters.
            location: The exact location to dispense to.
            well_core: The well to dispense to, if applicable.
            rate: The rate in µL/s to dispense at.
            flow_rate: Not used in this core.
            in_place: Whether we should move_to location.
            push_out: The amount to push the plunger below bottom position.
        """
        if isinstance(location, (TrashBin, WasteChute)):
            raise APIVersionError(
                "Dispense in Moveable Trash or Waste Chute are not supported in this API Version."
            )
        if push_out:
            raise APIVersionError("push_out is not supported in this API version.")
        if not in_place:
            self.move_to(location=location)

        self._protocol_interface.get_hardware().dispense(self._mount, volume, rate)

    def blow_out(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[LegacyWellCore],
        in_place: bool,
    ) -> None:
        """Blow liquid out of the tip.

        Args:
            location: The location to blow out into.
            well_core: Unused by legacy core.
            in_place: Whether we should move_to location.
        """
        if isinstance(location, (TrashBin, WasteChute)):
            raise APIVersionError(
                "Blow Out in Moveable Trash or Waste Chute are not supported in this API Version."
            )

        if not in_place:
            self.move_to(location=location)
        self._protocol_interface.get_hardware().blow_out(self._mount)

    def touch_tip(
        self,
        location: types.Location,
        well_core: LegacyWellCore,
        radius: float,
        z_offset: float,
        speed: float,
    ) -> None:
        """
        Touch the pipette tip to the sides of a well, with the intent of
        removing left-over droplets
        """
        # TODO al 20201110 - build_edges relies on where being a Well. This is
        #  an unpleasant compromise until refactoring build_edges to support
        #  LegacyWellCore.
        #  Also, build_edges should not require api_version.
        from opentrons.protocol_api.labware import Labware, Well

        self.move_to(location=location)

        edges = build_edges(
            where=Well(
                parent=Labware(
                    core=well_core.geometry.parent,
                    api_version=self._api_version,
                    protocol_core=None,  # type: ignore[arg-type]
                    core_map=None,  # type: ignore[arg-type]
                ),
                core=well_core,
                api_version=self._api_version,
            ),
            offset=z_offset,
            mount=self._mount,
            deck=self._protocol_interface.get_deck(),
            radius=radius,
            version=self._api_version,
        )
        for edge in edges:
            self._protocol_interface.get_hardware().move_to(self._mount, edge, speed)

    def pick_up_tip(
        self,
        location: types.Location,
        well_core: LegacyWellCore,
        presses: Optional[int],
        increment: Optional[float],
        prep_after: bool = True,
    ) -> None:
        """Pick up a tip for the pipette to run liquid-handling commands."""
        hw = self._protocol_interface.get_hardware()
        geometry = well_core.geometry
        tip_rack_core = geometry.parent
        tip_length = instrument_support.tip_length_for(
            pipette=self.get_hardware_state(),
            tip_rack_definition=tip_rack_core.get_definition(),
        )

        self.move_to(location=location, well_core=well_core)
        hw.set_current_tiprack_diameter(self._mount, geometry.diameter)
        hw.pick_up_tip(self._mount, tip_length, presses, increment, prep_after)
        hw.set_working_volume(self._mount, geometry.max_volume)
        tip_rack_core.get_tip_tracker().use_tips(
            start_well=well_core,
            num_channels=self.get_channels(),
            fail_if_full=self._api_version < APIVersion(2, 2),
        )

    def drop_tip(
        self,
        location: Optional[types.Location],
        well_core: LegacyWellCore,
        home_after: Optional[bool],
        alternate_drop_location: Optional[bool] = False,
    ) -> None:
        """Move to and drop a tip into a given well.

        Args:
            location: An absolute location to drop the tip at.
                If unspecified, use the default drop height of the well.
            well_core: The well we're dropping into
            home_after: Whether to home the pipette after the tip is dropped.
        """
        if alternate_drop_location:
            raise APIVersionError(
                "Tip drop randomization is not supported in this API version."
            )
        labware_core = well_core.geometry.parent

        if location is None:
            from opentrons.protocol_api.labware import Labware, Well

            labware = Labware(
                core=labware_core,
                api_version=self._api_version,
                protocol_core=None,  # type: ignore[arg-type]
                core_map=None,  # type: ignore[arg-type]
            )
            well = Well(parent=labware, core=well_core, api_version=self._api_version)

            if LabwareLike(labware).is_fixed_trash():
                location = well.top()
            elif self._api_version < APIVersion(2, 2):
                location = well.bottom(z=_PRE_2_2_TIP_DROP_HEIGHT_MM)
            else:
                assert (
                    labware_core.is_tip_rack()
                ), "Expected tip drop target to be a tip rack."

                return_height = self.get_return_height()
                location = well.top(z=-return_height * labware_core.get_tip_length())

        hw = self._protocol_interface.get_hardware()
        self.move_to(location=location)
        hw.drop_tip(self._mount, home_after=True if home_after is None else home_after)

        if self._api_version < APIVersion(2, 2) and labware_core.is_tip_rack():
            # If this is a tiprack we can try and add the dirty tip back to the tracker
            try:
                labware_core.get_tip_tracker().return_tips(
                    start_well=well_core,
                    num_channels=self.get_channels(),
                )
            except AssertionError:
                # Similarly to :py:meth:`return_tips`, the failure case here
                # just means the tip can't be reused, so don't actually stop
                # the protocol
                _log.warning(
                    f"Could not return tip to {labware_core.get_display_name()}"
                )

    def drop_tip_in_disposal_location(
        self, disposal_location: Union[TrashBin, WasteChute], home_after: Optional[bool]
    ) -> None:
        raise APIVersionError(
            "Dropping tips in a trash bin or waste chute is not supported in this API Version."
        )

    def home(self) -> None:
        """Home the mount"""
        self._protocol_interface.get_hardware().home_z(
            mount=self._mount,
            # preserve buggy behavior in <= 2.12 for strict back. compat.
            # https://github.com/Opentrons/opentrons/issues/7499
            allow_home_other=self._api_version >= APIVersion(2, 13),
        )
        self.home_plunger()

    def home_plunger(self) -> None:
        """Home the plunger associated with this mount."""
        self._protocol_interface.get_hardware().home_plunger(self._mount)

    def move_to(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[LegacyWellCore] = None,
        force_direct: bool = False,
        minimum_z_height: Optional[float] = None,
        speed: Optional[float] = None,
    ) -> None:
        """Move the instrument.

        Args:
            location: The movement destination.
            well_core: Unused in the legacy instrument core.
            force_direct: Force a direct movement instead of an arc.
            minimum_z_height: Set a minimum travel height for a movement arc.
            speed: Override the travel speed in mm/s.

        Raises:
            LabwareHeightError: An item on the deck is taller than
                the computed safe travel height.
        """
        if isinstance(location, (TrashBin, WasteChute)):
            raise APIVersionError(
                "Move To Trash Bin and Waste Chute are not supported in this API Version."
            )
        self.flag_unsafe_move(location)

        # prevent direct movement bugs in PAPI version >= 2.10
        location_cache_mount = (
            self._mount if self._api_version >= APIVersion(2, 10) else None
        )

        last_location = self._protocol_interface.get_last_location(
            mount=location_cache_mount
        )

        if last_location:
            from_lw = last_location.labware
        else:
            from_lw = LabwareLike(None)

        if not speed:
            speed = self.get_default_speed()

        hardware = self._protocol_interface.get_hardware()

        from_center = from_lw.center_multichannel_on_wells() if from_lw else False
        cp_override = CriticalPoint.XY_CENTER if from_center else None

        from_loc = types.Location(
            hardware.gantry_position(self._mount, critical_point=cp_override), from_lw
        )

        instr_max_height = hardware.get_instrument_max_height(self._mount)
        moves = planning.plan_moves(
            from_loc,
            location,
            self._protocol_interface.get_deck(),
            instr_max_height,
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
        )

        try:
            for move in moves:
                hardware.move_to(
                    self._mount,
                    move[0],
                    critical_point=move[1],
                    speed=speed,
                    max_speeds=self._protocol_interface.get_max_speeds().data,
                )
        except Exception:
            self._protocol_interface.set_last_location(None)
            raise
        else:
            self._protocol_interface.set_last_location(
                location=location, mount=location_cache_mount
            )

    def get_mount(self) -> types.Mount:
        """Get the mount this pipette is attached to."""
        return self._mount

    def get_instrument_name(self) -> str:
        """Get the instrument name."""
        return self._instrument_name

    def get_pipette_name(self) -> str:
        """Get the pipette name."""
        return self.get_hardware_state()["name"]

    def get_model(self) -> str:
        """Get the model name."""
        return self.get_hardware_state()["model"]

    def get_display_name(self) -> str:
        """Get the display name"""
        return self.get_hardware_state()["display_name"]

    def get_min_volume(self) -> float:
        """Get the min volume."""
        return self.get_hardware_state()["min_volume"]

    def get_max_volume(self) -> float:
        """Get the max volume."""
        return self.get_hardware_state()["max_volume"]

    def get_working_volume(self) -> float:
        """Get the working volume."""
        return self.get_hardware_state()["working_volume"]

    def get_current_volume(self) -> float:
        """Get the current volume."""
        return self.get_hardware_state()["current_volume"]

    def get_available_volume(self) -> float:
        """Get the available volume."""
        return self.get_hardware_state()["available_volume"]

    def get_hardware_state(self) -> PipetteDict:
        """Get the hardware pipette dictionary."""
        sync_hw_api = self._protocol_interface.get_hardware()
        pipette: Optional[PipetteDict] = sync_hw_api.get_attached_instrument(
            self._mount
        )

        if pipette is None:
            raise types.PipetteNotAttachedError()

        return pipette

    def get_channels(self) -> int:
        """Number of channels."""
        return self.get_hardware_state()["channels"]

    def has_tip(self) -> bool:
        """Whether a tip is attached."""
        return self.get_hardware_state()["has_tip"]

    def is_ready_to_aspirate(self) -> bool:
        return self.get_hardware_state()["ready_to_aspirate"]

    def prepare_to_aspirate(self) -> None:
        self._protocol_interface.get_hardware().prepare_for_aspirate(self._mount)

    def get_return_height(self) -> float:
        """The height to return a tip to its tiprack."""
        return self.get_hardware_state().get("return_tip_height", 0.5)

    def get_flow_rate(self) -> FlowRates:
        return self._flow_rates

    def get_aspirate_flow_rate(self, rate: float = 1.0) -> float:
        return self.get_hardware_state()["aspirate_flow_rate"] * rate

    def get_dispense_flow_rate(self, rate: float = 1.0) -> float:
        return self.get_hardware_state()["dispense_flow_rate"] * rate

    def get_blow_out_flow_rate(self, rate: float = 1.0) -> float:
        return self.get_hardware_state()["blow_out_flow_rate"] * rate

    def get_speed(self) -> PlungerSpeeds:
        return self._speeds

    def set_flow_rate(
        self,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        """Set the flow rates."""
        self._protocol_interface.get_hardware().set_flow_rate(
            mount=self._mount,
            aspirate=aspirate,
            dispense=dispense,
            blow_out=blow_out,
        )

    def set_pipette_speed(
        self,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        """Set pipette speeds."""
        self._protocol_interface.get_hardware().set_pipette_speed(
            mount=self._mount,
            aspirate=aspirate,
            dispense=dispense,
            blow_out=blow_out,
        )

    def flag_unsafe_move(self, location: types.Location) -> None:
        """Check if a movement to a destination is potentially unsafe.

        Args:
            location: The movement destination.

        Raises:
            RuntimeError: The movement is unsafe.
        """
        from_loc = self._protocol_interface.get_last_location()

        if not from_loc:
            from_loc = types.Location(types.Point(0, 0, 0), LabwareLike(None))

        for mod in self._protocol_interface.get_module_cores():
            if isinstance(mod, LegacyThermocyclerCore):
                mod.flag_unsafe_move(to_loc=location, from_loc=from_loc)
            elif isinstance(mod, LegacyHeaterShakerCore):
                mod.flag_unsafe_move(
                    to_loc=location,
                    is_multichannel=self.get_channels() > 1,
                )

    def configure_for_volume(self, volume: float) -> None:
        """This will never be called because it was added in API 2.15."""
        pass

    def configure_nozzle_layout(
        self,
        style: NozzleLayout,
        primary_nozzle: Optional[str],
        front_right_nozzle: Optional[str],
    ) -> None:
        """This will never be called because it was added in API 2.16."""
        pass

    def get_active_channels(self) -> int:
        """This will never be called because it was added in API 2.16."""
        assert False, "get_active_channels only supported in API 2.16 & later"

    def is_tip_tracking_available(self) -> bool:
        # Tip tracking is always available in legacy context
        return True



"""Public protocol engine value types and models."""
from __future__ import annotations
import re
from datetime import datetime
from enum import Enum
from dataclasses import dataclass
from pydantic import BaseModel, Field, validator
from typing import Optional, Union, List, Dict, Any, NamedTuple, Tuple, FrozenSet
from typing_extensions import Literal, TypeGuard

from opentrons_shared_data.pipette.dev_types import PipetteNameType
from opentrons.types import MountType, DeckSlotName, StagingSlotName
from opentrons.hardware_control.types import TipStateType as HwTipStateType
from opentrons.hardware_control.modules import (
    ModuleType as ModuleType,
)

from opentrons_shared_data.pipette.dev_types import (  # noqa: F401
    # convenience re-export of LabwareUri type
    LabwareUri as LabwareUri,
)
from opentrons_shared_data.module.dev_types import ModuleType as SharedDataModuleType


class EngineStatus(str, Enum):
    """Current execution status of a ProtocolEngine."""

    IDLE = "idle"
    RUNNING = "running"
    PAUSED = "paused"
    BLOCKED_BY_OPEN_DOOR = "blocked-by-open-door"
    STOP_REQUESTED = "stop-requested"
    STOPPED = "stopped"
    FINISHING = "finishing"
    FAILED = "failed"
    SUCCEEDED = "succeeded"


class DeckSlotLocation(BaseModel):
    """The location of something placed in a single deck slot."""

    slotName: DeckSlotName = Field(
        ...,
        description=(
            # This description should be kept in sync with LabwareOffsetLocation.slotName.
            "A slot on the robot's deck."
            "\n\n"
            'The plain numbers like `"5"` are for the OT-2,'
            ' and the coordinates like `"C2"` are for the Flex.'
            "\n\n"
            "When you provide one of these values, you can use either style."
            " It will automatically be converted to match the robot."
            "\n\n"
            "When one of these values is returned, it will always match the robot."
        ),
    )


class StagingSlotLocation(BaseModel):
    """The location of something placed in a single staging slot."""

    slotName: StagingSlotName = Field(
        ...,
        description=(
            # This description should be kept in sync with LabwareOffsetLocation.slotName.
            "A slot on the robot's staging area."
            "\n\n"
            "These apply only to the Flex. The OT-2 has no staging slots."
        ),
    )


class AddressableAreaLocation(BaseModel):
    """The location of something place in an addressable area. This is a superset of deck slots."""

    addressableAreaName: str = Field(
        ...,
        description=(
            "The name of the addressable area that you want to use."
            " Valid values are the `id`s of `addressableArea`s in the"
            " [deck definition](https://github.com/Opentrons/opentrons/tree/edge/shared-data/deck)."
        ),
    )


class ModuleLocation(BaseModel):
    """The location of something placed atop a hardware module."""

    moduleId: str = Field(
        ...,
        description="The ID of a loaded module from a prior `loadModule` command.",
    )


class OnLabwareLocation(BaseModel):
    """The location of something placed atop another labware."""

    labwareId: str = Field(
        ...,
        description="The ID of a loaded Labware from a prior `loadLabware` command.",
    )


_OffDeckLocationType = Literal["offDeck"]
OFF_DECK_LOCATION: _OffDeckLocationType = "offDeck"

LabwareLocation = Union[
    DeckSlotLocation,
    ModuleLocation,
    OnLabwareLocation,
    _OffDeckLocationType,
    AddressableAreaLocation,
]
"""Union of all locations where it's legal to keep a labware."""

OnDeckLabwareLocation = Union[
    DeckSlotLocation, ModuleLocation, OnLabwareLocation, AddressableAreaLocation
]

NonStackedLocation = Union[
    DeckSlotLocation, AddressableAreaLocation, ModuleLocation, _OffDeckLocationType
]
"""Union of all locations where it's legal to keep a labware that can't be stacked on another labware"""


class WellOrigin(str, Enum):
    """Origin of WellLocation offset.

    Props:
        TOP: the top-center of the well
        BOTTOM: the bottom-center of the well
        CENTER: the middle-center of the well
    """

    TOP = "top"
    BOTTOM = "bottom"
    CENTER = "center"


class DropTipWellOrigin(str, Enum):
    """The origin of a DropTipWellLocation offset.

    Props:
        TOP: the top-center of the well
        BOTTOM: the bottom-center of the well
        CENTER: the middle-center of the well
        DEFAULT: the default drop-tip location of the well,
            based on pipette configuration and length of the tip.
    """

    TOP = "top"
    BOTTOM = "bottom"
    CENTER = "center"
    DEFAULT = "default"


# This is deliberately a separate type from Vec3f to let components default to 0.
class WellOffset(BaseModel):
    """An offset vector in (x, y, z)."""

    x: float = 0
    y: float = 0
    z: float = 0


class WellLocation(BaseModel):
    """A relative location in reference to a well's location."""

    origin: WellOrigin = WellOrigin.TOP
    offset: WellOffset = Field(default_factory=WellOffset)


class DropTipWellLocation(BaseModel):
    """Like WellLocation, but for dropping tips.

    Unlike a typical WellLocation, the location for a drop tip
    defaults to location based on the tip length rather than the well's top.
    """

    origin: DropTipWellOrigin = DropTipWellOrigin.DEFAULT
    offset: WellOffset = Field(default_factory=WellOffset)


@dataclass(frozen=True)
class Dimensions:
    """Dimensions of an object in deck-space."""

    x: float
    y: float
    z: float


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class DeckPoint(BaseModel):
    """Coordinates of a point in deck space."""

    x: float
    y: float
    z: float


# TODO(mm, 2023-05-10): Deduplicate with constants in
# opentrons.protocols.api_support.deck_type
# and consider moving to shared-data.
class DeckType(str, Enum):
    """Types of deck available."""

    OT2_STANDARD = "ot2_standard"
    OT2_SHORT_TRASH = "ot2_short_trash"
    OT3_STANDARD = "ot3_standard"


class LoadedPipette(BaseModel):
    """A pipette that has been loaded."""

    id: str
    pipetteName: PipetteNameType
    mount: MountType


@dataclass
class FlowRates:
    """Default and current flow rates for a pipette."""

    default_blow_out: Dict[str, float]
    default_aspirate: Dict[str, float]
    default_dispense: Dict[str, float]


@dataclass(frozen=True)
class CurrentWell:
    """The latest well that the robot has accessed."""

    pipette_id: str
    labware_id: str
    well_name: str


@dataclass(frozen=True)
class CurrentAddressableArea:
    """The latest addressable area the robot has accessed."""

    pipette_id: str
    addressable_area_name: str


CurrentPipetteLocation = Union[CurrentWell, CurrentAddressableArea]


@dataclass(frozen=True)
class TipGeometry:
    """Tip geometry data.

    Props:
        length: The effective length (total length minus overlap) of a tip in mm.
        diameter: Tip diameter in mm.
        volume: Maximum volume in µL.
    """

    length: float
    diameter: float
    volume: float


class MovementAxis(str, Enum):
    """Axis on which to issue a relative movement."""

    X = "x"
    Y = "y"
    Z = "z"


class MotorAxis(str, Enum):
    """Motor axis on which to issue a home command."""

    X = "x"
    Y = "y"
    LEFT_Z = "leftZ"
    RIGHT_Z = "rightZ"
    LEFT_PLUNGER = "leftPlunger"
    RIGHT_PLUNGER = "rightPlunger"
    EXTENSION_Z = "extensionZ"
    EXTENSION_JAW = "extensionJaw"


# TODO(mc, 2022-01-18): use opentrons_shared_data.module.dev_types.ModuleModel
class ModuleModel(str, Enum):
    """All available modules' models."""

    TEMPERATURE_MODULE_V1 = "temperatureModuleV1"
    TEMPERATURE_MODULE_V2 = "temperatureModuleV2"
    MAGNETIC_MODULE_V1 = "magneticModuleV1"
    MAGNETIC_MODULE_V2 = "magneticModuleV2"
    THERMOCYCLER_MODULE_V1 = "thermocyclerModuleV1"
    THERMOCYCLER_MODULE_V2 = "thermocyclerModuleV2"
    HEATER_SHAKER_MODULE_V1 = "heaterShakerModuleV1"
    MAGNETIC_BLOCK_V1 = "magneticBlockV1"

    def as_type(self) -> ModuleType:
        """Get the ModuleType of this model."""
        if ModuleModel.is_temperature_module_model(self):
            return ModuleType.TEMPERATURE
        elif ModuleModel.is_magnetic_module_model(self):
            return ModuleType.MAGNETIC
        elif ModuleModel.is_thermocycler_module_model(self):
            return ModuleType.THERMOCYCLER
        elif ModuleModel.is_heater_shaker_module_model(self):
            return ModuleType.HEATER_SHAKER
        elif ModuleModel.is_magnetic_block(self):
            return ModuleType.MAGNETIC_BLOCK

        assert False, f"Invalid ModuleModel {self}"

    @classmethod
    def is_temperature_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[TemperatureModuleModel]:
        """Whether a given model is a Temperature Module."""
        return model in [cls.TEMPERATURE_MODULE_V1, cls.TEMPERATURE_MODULE_V2]

    @classmethod
    def is_magnetic_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[MagneticModuleModel]:
        """Whether a given model is a Magnetic Module."""
        return model in [cls.MAGNETIC_MODULE_V1, cls.MAGNETIC_MODULE_V2]

    @classmethod
    def is_thermocycler_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[ThermocyclerModuleModel]:
        """Whether a given model is a Thermocycler Module."""
        return model in [cls.THERMOCYCLER_MODULE_V1, cls.THERMOCYCLER_MODULE_V2]

    @classmethod
    def is_heater_shaker_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[HeaterShakerModuleModel]:
        """Whether a given model is a Heater-Shaker Module."""
        return model == cls.HEATER_SHAKER_MODULE_V1

    @classmethod
    def is_magnetic_block(cls, model: ModuleModel) -> TypeGuard[MagneticBlockModel]:
        """Whether a given model is a Magnetic block."""
        return model == cls.MAGNETIC_BLOCK_V1


TemperatureModuleModel = Literal[
    ModuleModel.TEMPERATURE_MODULE_V1, ModuleModel.TEMPERATURE_MODULE_V2
]
MagneticModuleModel = Literal[
    ModuleModel.MAGNETIC_MODULE_V1, ModuleModel.MAGNETIC_MODULE_V2
]
ThermocyclerModuleModel = Literal[
    ModuleModel.THERMOCYCLER_MODULE_V1, ModuleModel.THERMOCYCLER_MODULE_V2
]
HeaterShakerModuleModel = Literal[ModuleModel.HEATER_SHAKER_MODULE_V1]
MagneticBlockModel = Literal[ModuleModel.MAGNETIC_BLOCK_V1]


class ModuleDimensions(BaseModel):
    """Dimension type for modules."""

    bareOverallHeight: float
    overLabwareHeight: float
    lidHeight: Optional[float]


class Vec3f(BaseModel):
    """A 3D vector of floats."""

    x: float
    y: float
    z: float


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class ModuleCalibrationPoint(BaseModel):
    """Calibration Point type for module definition."""

    x: float
    y: float
    z: float


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class LabwareOffsetVector(BaseModel):
    """Offset, in deck coordinates from nominal to actual position."""

    x: float
    y: float
    z: float

    def __add__(self, other: Any) -> LabwareOffsetVector:
        """Adds two vectors together."""
        if not isinstance(other, LabwareOffsetVector):
            return NotImplemented
        return LabwareOffsetVector(
            x=self.x + other.x, y=self.y + other.y, z=self.z + other.z
        )

    def __sub__(self, other: Any) -> LabwareOffsetVector:
        """Subtracts two vectors."""
        if not isinstance(other, LabwareOffsetVector):
            return NotImplemented
        return LabwareOffsetVector(
            x=self.x - other.x, y=self.y - other.y, z=self.z - other.z
        )


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class InstrumentOffsetVector(BaseModel):
    """Instrument Offset from home position to robot deck."""

    x: float
    y: float
    z: float


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class ModuleOffsetVector(BaseModel):
    """Offset, in deck coordinates, from nominal to actual position of labware on a module."""

    x: float
    y: float
    z: float


@dataclass
class ModuleOffsetData:
    """Module calibration offset data."""

    moduleOffsetVector: ModuleOffsetVector
    location: DeckSlotLocation


class OverlapOffset(Vec3f):
    """Offset representing overlap space of one labware on top of another labware or module."""


class AddressableOffsetVector(Vec3f):
    """Offset, in deck coordinates, from nominal to actual position of an addressable area."""


class LabwareMovementOffsetData(BaseModel):
    """Offsets to be used during labware movement."""

    pickUpOffset: LabwareOffsetVector
    dropOffset: LabwareOffsetVector


# TODO(mm, 2023-04-13): Move to shared-data, so this binding can be maintained alongside the JSON
# schema that it's sourced from. We already do that for labware definitions and JSON protocols.
class ModuleDefinition(BaseModel):
    """A module definition conforming to module definition schema v3."""

    # Note: This field is misleading.
    #
    # This class only models v3 definitions ("module/schemas/3"), not v2 ("module/schemas/2").
    # labwareOffset is required to have a z-component, for example.
    #
    # When parsing from a schema v3 JSON definition into this model,
    # the definition's `"$otSharedSchema": "module/schemas/3"` field will be thrown away
    # because it has a dollar sign, which doesn't match this field.
    # Then, this field will default to "module/schemas/2", because no value was provided.
    #
    # We should fix this field once Jira RSS-221 is resolved. RSS-221 makes it difficult to fix
    # because robot-server has been storing and loading these bad fields in its database.
    otSharedSchema: str = Field("module/schemas/2", description="The current schema.")

    moduleType: ModuleType = Field(
        ...,
        description="Module type (Temperature/Magnetic/Thermocycler)",
    )

    model: ModuleModel = Field(..., description="Model name of the module")

    labwareOffset: LabwareOffsetVector = Field(
        ...,
        description="Labware offset in x, y, z.",
    )

    dimensions: ModuleDimensions = Field(..., description="Module dimension")

    calibrationPoint: ModuleCalibrationPoint = Field(
        ...,
        description="Calibration point of module.",
    )

    displayName: str = Field(..., description="Display name.")

    quirks: List[str] = Field(..., description="Module quirks")

    # In releases prior to https://github.com/Opentrons/opentrons/pull/11873 (v6.3.0),
    # the matrices in slotTransforms were 3x3.
    # After, they are 4x4, even though there was no schema version bump.
    #
    # Because old objects of this class, with the 3x3 matrices, were stored in robot-server's
    # database, this field needs to stay typed loosely enough to support both sizes.
    # We can fix this once Jira RSS-221 is resolved.
    slotTransforms: Dict[str, Any] = Field(
        ...,
        description="Dictionary of transforms for each slot.",
    )

    compatibleWith: List[ModuleModel] = Field(
        ...,
        description="List of module models this model is compatible with.",
    )
    gripperOffsets: Optional[Dict[str, LabwareMovementOffsetData]] = Field(
        default_factory=dict,
        description="Offsets to use for labware movement using gripper",
    )


class LoadedModule(BaseModel):
    """A module that has been loaded."""

    id: str
    model: ModuleModel
    location: Optional[DeckSlotLocation]
    serialNumber: Optional[str]


class LabwareOffsetLocation(BaseModel):
    """Parameters describing when a given offset may apply to a given labware load."""

    slotName: DeckSlotName = Field(
        ...,
        description=(
            "The deck slot where the protocol will load the labware."
            " Or, if the protocol will load the labware on a module,"
            " the deck slot where the protocol will load that module."
            "\n\n"
            # This description should be kept in sync with DeckSlotLocation.slotName.
            'The plain numbers like `"5"` are for the OT-2,'
            ' and the coordinates like `"C2"` are for the Flex.'
            "\n\n"
            "When you provide one of these values, you can use either style."
            " It will automatically be converted to match the robot."
            "\n\n"
            "When one of these values is returned, it will always match the robot."
        ),
    )
    moduleModel: Optional[ModuleModel] = Field(
        None,
        description=(
            "The model of the module that the labware will be loaded onto,"
            " if applicable."
            "\n\n"
            "Because of module compatibility, the model that the protocol requests"
            " may not be exactly the same"
            " as what it will find physically connected during execution."
            " For this labware offset to apply,"
            " this field must be the *requested* model, not the connected one."
            " You can retrieve this from a `loadModule` command's `params.model`"
            " in the protocol's analysis."
        ),
    )
    definitionUri: Optional[str] = Field(
        None,
        description=(
            "The definition URI of a labware that a labware can be loaded onto,"
            " if applicable."
            "\n\n"
            "This can be combined with moduleModel if the labware is loaded on top of"
            " an adapter that is loaded on a module."
        ),
    )


class LabwareOffset(BaseModel):
    """An offset that the robot adds to a pipette's position when it moves to a labware.

    During the run, if a labware is loaded whose definition URI and location
    both match what's found here, the given offset will be added to all
    pipette movements that use that labware as a reference point.
    """

    id: str = Field(..., description="Unique labware offset record identifier.")
    createdAt: datetime = Field(..., description="When this labware offset was added.")
    definitionUri: str = Field(..., description="The URI for the labware's definition.")
    location: LabwareOffsetLocation = Field(
        ...,
        description="Where the labware is located on the robot.",
    )
    vector: LabwareOffsetVector = Field(
        ...,
        description="The offset applied to matching labware.",
    )


class LabwareOffsetCreate(BaseModel):
    """Create request data for a labware offset."""

    definitionUri: str = Field(..., description="The URI for the labware's definition.")
    location: LabwareOffsetLocation = Field(
        ...,
        description="Where the labware is located on the robot.",
    )
    vector: LabwareOffsetVector = Field(
        ...,
        description="The offset applied to matching labware.",
    )


class LoadedLabware(BaseModel):
    """A labware that has been loaded."""

    id: str
    loadName: str
    definitionUri: str
    location: LabwareLocation = Field(
        ..., description="The labware's current location."
    )
    offsetId: Optional[str] = Field(
        None,
        description=(
            "An ID referencing the labware offset"
            " that applies to this labware placement."
            " Null or undefined means no offset was provided for this load,"
            " so the default of (0, 0, 0) will be used."
        ),
    )
    displayName: Optional[str] = Field(
        None,
        description="A user-specified display name for this labware, if provided.",
    )


class HexColor(BaseModel):
    """Hex color representation."""

    __root__: str

    @validator("__root__")
    def _color_is_a_valid_hex(cls, v: str) -> str:
        match = re.search(r"^#(?:[0-9a-fA-F]{3,4}){1,2}$", v)
        if not match:
            raise ValueError("Color is not a valid hex color.")
        return v


class Liquid(BaseModel):
    """Payload required to create a liquid."""

    id: str
    displayName: str
    description: str
    displayColor: Optional[HexColor]


class SpeedRange(NamedTuple):
    """Minimum and maximum allowed speeds for a shaking module."""

    min: int
    max: int


class TemperatureRange(NamedTuple):
    """Minimum and maximum allowed temperatures for a heating module."""

    min: float
    max: float


class HeaterShakerLatchStatus(Enum):
    """Heater-Shaker latch status for determining pipette and labware movement errors."""

    CLOSED = "closed"
    OPEN = "open"
    UNKNOWN = "unknown"


@dataclass(frozen=True)
class HeaterShakerMovementRestrictors:
    """Shaking status, latch status and slot location for determining movement restrictions."""

    plate_shaking: bool
    latch_status: HeaterShakerLatchStatus
    deck_slot: int


class LabwareMovementStrategy(str, Enum):
    """Strategy to use for labware movement."""

    USING_GRIPPER = "usingGripper"
    MANUAL_MOVE_WITH_PAUSE = "manualMoveWithPause"
    MANUAL_MOVE_WITHOUT_PAUSE = "manualMoveWithoutPause"


@dataclass(frozen=True)
class PotentialCutoutFixture:
    """Cutout and cutout fixture id associated with a potential cutout fixture that can be on the deck."""

    cutout_id: str
    cutout_fixture_id: str
    provided_addressable_areas: FrozenSet[str]


class AreaType(Enum):
    """The type of addressable area."""

    SLOT = "slot"
    STAGING_SLOT = "stagingSlot"
    MOVABLE_TRASH = "movableTrash"
    FIXED_TRASH = "fixedTrash"
    WASTE_CHUTE = "wasteChute"


@dataclass(frozen=True)
class AddressableArea:
    """Addressable area that has been loaded."""

    area_name: str
    area_type: AreaType
    base_slot: DeckSlotName
    display_name: str
    bounding_box: Dimensions
    position: AddressableOffsetVector
    compatible_module_types: List[SharedDataModuleType]


class PostRunHardwareState(Enum):
    """State of robot gantry & motors after a stop is performed and the hardware API is reset.

    HOME_AND_STAY_ENGAGED: home the gantry and keep all motors engaged. This allows the
        robot to continue performing movement actions without re-homing
    HOME_THEN_DISENGAGE: home the gantry and then disengage motors.
        Reduces current consumption of the motors and prevents coil heating.
        Re-homing is required to re-engage the motors and resume robot movement.
    STAY_ENGAGED_IN_PLACE: do not home after the stop and keep the motors engaged.
        Keeps gantry in the same position as prior to `stop()` execution
        and allows the robot to execute movement commands without requiring to re-home first.
    DISENGAGE_IN_PLACE: disengage motors and do not home the robot
    Probable states for pipette:
        - for 1- or 8-channel:
            - HOME_AND_STAY_ENGAGED after protocol runs
            - STAY_ENGAGED_IN_PLACE after maintenance runs
        - for 96-channel:
            - HOME_THEN_DISENGAGE after protocol runs
            - DISENGAGE_IN_PLACE after maintenance runs
    """

    HOME_AND_STAY_ENGAGED = "homeAndStayEngaged"
    HOME_THEN_DISENGAGE = "homeThenDisengage"
    STAY_ENGAGED_IN_PLACE = "stayEngagedInPlace"
    DISENGAGE_IN_PLACE = "disengageInPlace"


NOZZLE_NAME_REGEX = "[A-Z][0-100]"
PRIMARY_NOZZLE_LITERAL = Literal["A1", "H1", "A12", "H12"]


class AllNozzleLayoutConfiguration(BaseModel):
    """All basemodel to represent a reset to the nozzle configuration. Sending no parameters resets to default."""

    style: Literal["ALL"] = "ALL"


class SingleNozzleLayoutConfiguration(BaseModel):
    """Minimum information required for a new nozzle configuration."""

    style: Literal["SINGLE"] = "SINGLE"
    primaryNozzle: PRIMARY_NOZZLE_LITERAL = Field(
        ...,
        description="The primary nozzle to use in the layout configuration. This nozzle will update the critical point of the current pipette. For now, this is also the back left corner of your rectangle.",
    )


class RowNozzleLayoutConfiguration(BaseModel):
    """Minimum information required for a new nozzle configuration."""

    style: Literal["ROW"] = "ROW"
    primaryNozzle: PRIMARY_NOZZLE_LITERAL = Field(
        ...,
        description="The primary nozzle to use in the layout configuration. This nozzle will update the critical point of the current pipette. For now, this is also the back left corner of your rectangle.",
    )


class ColumnNozzleLayoutConfiguration(BaseModel):
    """Information required for nozzle configurations of type ROW and COLUMN."""

    style: Literal["COLUMN"] = "COLUMN"
    primaryNozzle: PRIMARY_NOZZLE_LITERAL = Field(
        ...,
        description="The primary nozzle to use in the layout configuration. This nozzle will update the critical point of the current pipette. For now, this is also the back left corner of your rectangle.",
    )


class QuadrantNozzleLayoutConfiguration(BaseModel):
    """Information required for nozzle configurations of type QUADRANT."""

    style: Literal["QUADRANT"] = "QUADRANT"
    primaryNozzle: PRIMARY_NOZZLE_LITERAL = Field(
        ...,
        description="The primary nozzle to use in the layout configuration. This nozzle will update the critical point of the current pipette. For now, this is also the back left corner of your rectangle.",
    )
    frontRightNozzle: str = Field(
        ...,
        regex=NOZZLE_NAME_REGEX,
        description="The front right nozzle in your configuration.",
    )


NozzleLayoutConfigurationType = Union[
    AllNozzleLayoutConfiguration,
    SingleNozzleLayoutConfiguration,
    ColumnNozzleLayoutConfiguration,
    RowNozzleLayoutConfiguration,
    QuadrantNozzleLayoutConfiguration,
]

# TODO make the below some sort of better type
DeckConfigurationType = List[Tuple[str, str]]  # cutout_id, cutout_fixture_id


class TipPresenceStatus(str, Enum):
    """Tip presence status reported by a pipette."""

    PRESENT = "present"
    ABSENT = "absent"
    UNKNOWN = "unknown"

    def to_hw_state(self) -> HwTipStateType:
        """Convert to hardware tip state."""
        assert self != TipPresenceStatus.UNKNOWN
        return {
            TipPresenceStatus.PRESENT: HwTipStateType.PRESENT,
            TipPresenceStatus.ABSENT: HwTipStateType.ABSENT,
        }[self]

    @classmethod
    def from_hw_state(cls, state: HwTipStateType) -> "TipPresenceStatus":
        """Convert from hardware tip state."""
        return {
            HwTipStateType.PRESENT: TipPresenceStatus.PRESENT,
            HwTipStateType.ABSENT: TipPresenceStatus.ABSENT,
        }[state]



"""Motion planning module."""

from .waypoints import (
    DEFAULT_GENERAL_ARC_Z_MARGIN,
    DEFAULT_IN_LABWARE_ARC_Z_MARGIN,
    MINIMUM_Z_MARGIN,
    get_waypoints,
    get_gripper_labware_movement_waypoints,
)

from .types import Waypoint, MoveType

from .errors import (
    MotionPlanningError,
    DestinationOutOfBoundsError,
    ArcOutOfBoundsError,
)

__all__ = [
    "DEFAULT_GENERAL_ARC_Z_MARGIN",
    "DEFAULT_IN_LABWARE_ARC_Z_MARGIN",
    "MINIMUM_Z_MARGIN",
    "Waypoint",
    "MoveType",
    "MotionPlanningError",
    "DestinationOutOfBoundsError",
    "ArcOutOfBoundsError",
    "get_waypoints",
    "get_gripper_labware_movement_waypoints",
]



"""State store actions.

Actions can be passed to the ActionDispatcher, where they will trigger
reactions in objects that subscribe to the pipeline, like the StateStore.
"""
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Optional, Union

from opentrons.protocols.models import LabwareDefinition
from opentrons.hardware_control.types import DoorState
from opentrons.hardware_control.modules import LiveData

from opentrons_shared_data.errors import EnumeratedError

from ..commands import Command, CommandCreate, CommandPrivateResult
from ..types import (
    LabwareOffsetCreate,
    ModuleDefinition,
    Liquid,
    DeckConfigurationType,
    AddressableAreaLocation,
)


@dataclass(frozen=True)
class PlayAction:
    """Start or resume processing commands in the engine."""

    requested_at: datetime
    deck_configuration: Optional[DeckConfigurationType]


class PauseSource(str, Enum):
    """The source of a PauseAction.

    Attributes:
        CLIENT: the pause came externally, from the engine client.
        PROTOCOL: the pause came from the protocol itself.
    """

    CLIENT = "client"
    PROTOCOL = "protocol"


@dataclass(frozen=True)
class PauseAction:
    """Pause processing commands in the engine."""

    source: PauseSource


@dataclass(frozen=True)
class StopAction:
    """Stop the current engine execution.

    After a StopAction, the engine status will be marked as stopped.
    """

    from_estop: bool = False


@dataclass(frozen=True)
class FinishErrorDetails:
    """Error details for the payload of a FinishAction or HardwareStoppedAction."""

    error: Exception
    error_id: str
    created_at: datetime


@dataclass(frozen=True)
class FinishAction:
    """Gracefully stop processing commands in the engine."""

    set_run_status: bool = True
    """Whether to set the engine status depending on `error_details`.

    If True, the engine status will be marked `succeeded` or `failed`, depending on `error_details`.
    If False, the engine status will be marked `stopped`.
    """

    error_details: Optional[FinishErrorDetails] = None
    """The fatal error that caused the run to fail."""


@dataclass(frozen=True)
class HardwareStoppedAction:
    """An action dispatched after hardware has been stopped for good, for this engine instance."""

    completed_at: datetime

    finish_error_details: Optional[FinishErrorDetails]
    """The error that happened while doing post-run finish steps (homing and dropping tips)."""


@dataclass(frozen=True)
class DoorChangeAction:
    """Handle events coming in from hardware control."""

    door_state: DoorState


@dataclass(frozen=True)
class QueueCommandAction:
    """Add a command request to the queue."""

    command_id: str
    created_at: datetime
    request: CommandCreate
    request_hash: Optional[str]


@dataclass(frozen=True)
class UpdateCommandAction:
    """Update a given command."""

    command: Command
    private_result: CommandPrivateResult


@dataclass(frozen=True)
class FailCommandAction:
    """Mark a given command as failed.

    The given command and all currently queued commands will be marked
    as failed due to the given error.
    """

    # TODO(mc, 2021-11-12): we'll likely need to add the command params
    # to this payload for state reaction purposes
    command_id: str
    error_id: str
    failed_at: datetime
    error: EnumeratedError


@dataclass(frozen=True)
class AddLabwareOffsetAction:
    """Add a labware offset, to apply to subsequent `LoadLabwareCommand`s."""

    labware_offset_id: str
    created_at: datetime
    request: LabwareOffsetCreate


@dataclass(frozen=True)
class AddLabwareDefinitionAction:
    """Add a labware definition, to apply to subsequent `LoadLabwareCommand`s."""

    definition: LabwareDefinition


@dataclass(frozen=True)
class AddLiquidAction:
    """Add a liquid, to apply to subsequent `LoadLiquid`s."""

    liquid: Liquid


@dataclass(frozen=True)
class AddAddressableAreaAction:
    """Add a single addressable area to state.

    This differs from the deck configuration in PlayAction which sends over a mapping of cutout fixtures.
    This action will only load one addressable area and that should be pre-validated before being sent via
    the action.
    """

    addressable_area: AddressableAreaLocation


@dataclass(frozen=True)
class AddModuleAction:
    """Add an attached module directly to state without a location."""

    module_id: str
    serial_number: str
    definition: ModuleDefinition
    module_live_data: LiveData


@dataclass(frozen=True)
class ResetTipsAction:
    """Reset the tip tracking state of a given tip rack."""

    labware_id: str


@dataclass(frozen=True)
class SetPipetteMovementSpeedAction:
    """Set the speed of a pipette's X/Y/Z movements. Does not affect plunger speed.

    None will use the hardware API's default.
    """

    pipette_id: str
    speed: Optional[float]


Action = Union[
    PlayAction,
    PauseAction,
    StopAction,
    FinishAction,
    HardwareStoppedAction,
    DoorChangeAction,
    QueueCommandAction,
    UpdateCommandAction,
    FailCommandAction,
    AddLabwareOffsetAction,
    AddLabwareDefinitionAction,
    AddModuleAction,
    AddAddressableAreaAction,
    AddLiquidAction,
    ResetTipsAction,
    SetPipetteMovementSpeedAction,
]
